# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the photometry module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.nddata import NDData, StdDevUncertainty
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose, assert_array_less, assert_equal

from photutils.aperture.circle import (CircularAnnulus, CircularAperture,
                                       SkyCircularAnnulus, SkyCircularAperture)
from photutils.aperture.ellipse import (EllipticalAnnulus, EllipticalAperture,
                                        SkyEllipticalAnnulus,
                                        SkyEllipticalAperture)
from photutils.aperture.photometry import aperture_photometry
from photutils.aperture.rectangle import (RectangularAnnulus,
                                          RectangularAperture,
                                          SkyRectangularAnnulus,
                                          SkyRectangularAperture)
from photutils.datasets import make_4gaussians_image, make_gwcs, make_wcs
from photutils.utils._optional_deps import (HAS_GWCS, HAS_MATPLOTLIB,
                                            HAS_REGIONS)

APERTURE_CL = [CircularAperture,
               CircularAnnulus,
               EllipticalAperture,
               EllipticalAnnulus,
               RectangularAperture,
               RectangularAnnulus]

TEST_APERTURES = list(zip(APERTURE_CL,
                          ({'r': 3.0},
                           {'r_in': 3.0, 'r_out': 5.0},
                           {'a': 3.0, 'b': 5.0, 'theta': 1.0},
                           {'a_in': 3.0, 'a_out': 5.0, 'b_out': 4.0,
                            'b_in': 12.0 / 5.0, 'theta': 1.0},
                           {'w': 5, 'h': 8, 'theta': np.pi / 4},
                           {'w_in': 8, 'w_out': 12, 'h_out': 8,
                            'h_in': 16.0 / 3.0, 'theta': np.pi / 8}),
                          strict=True))


@pytest.mark.parametrize(('aperture_class', 'params'), TEST_APERTURES)
def test_outside_array(aperture_class, params):
    data = np.ones((10, 10), dtype=float)
    aperture = aperture_class((-60, 60), **params)
    fluxtable = aperture_photometry(data, aperture)
    # Aperture is fully outside array
    assert np.isnan(fluxtable['aperture_sum'])


@pytest.mark.parametrize(('aperture_class', 'params'), TEST_APERTURES)
def test_inside_array_simple(aperture_class, params):
    data = np.ones((40, 40), dtype=float)
    aperture = aperture_class((20.0, 20.0), **params)
    table1 = aperture_photometry(data, aperture, method='center',
                                 subpixels=10)
    table2 = aperture_photometry(data, aperture, method='subpixel',
                                 subpixels=10)
    table3 = aperture_photometry(data, aperture, method='exact', subpixels=10)
    true_flux = aperture.area
    assert table1['aperture_sum'] < table3['aperture_sum']

    if not isinstance(aperture, (RectangularAperture, RectangularAnnulus)):
        assert_allclose(table3['aperture_sum'], true_flux)
        assert_allclose(table2['aperture_sum'], table3['aperture_sum'],
                        atol=0.1)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
@pytest.mark.parametrize(('aperture_class', 'params'), TEST_APERTURES)
def test_aperture_plots(aperture_class, params):
    # This test should run without any errors, and there is no return
    # value.
    aperture = aperture_class((20.0, 20.0), **params)
    aperture.plot()


def test_aperture_pixel_positions():
    pos1 = (10, 20)
    pos2 = [(10, 20)]
    r = 3
    ap1 = CircularAperture(pos1, r)
    ap2 = CircularAperture(pos2, r)
    assert not np.array_equal(ap1.positions, ap2.positions)


class BaseTestAperturePhotometry:
    def test_array_error(self):
        # Array error
        error = np.ones(self.data.shape, dtype=float)
        if not hasattr(self, 'mask'):
            mask = None
            true_error = np.sqrt(self.area)
        else:
            mask = self.mask
            # 1 masked pixel
            true_error = np.sqrt(self.area - 1)

        table1 = aperture_photometry(self.data,
                                     self.aperture, method='center',
                                     mask=mask, error=error)
        table2 = aperture_photometry(self.data,
                                     self.aperture,
                                     method='subpixel', subpixels=12,
                                     mask=mask, error=error)
        table3 = aperture_photometry(self.data,
                                     self.aperture, method='exact',
                                     mask=mask, error=error)

        if not isinstance(self.aperture, (RectangularAperture,
                                          RectangularAnnulus)):
            assert_allclose(table3['aperture_sum'], self.true_flux)
            assert_allclose(table2['aperture_sum'], table3['aperture_sum'],
                            atol=0.1)
        assert np.all(table1['aperture_sum'] < table3['aperture_sum'])

        if not isinstance(self.aperture, (RectangularAperture,
                                          RectangularAnnulus)):
            assert_allclose(table3['aperture_sum_err'], true_error)
            assert_allclose(table2['aperture_sum_err'],
                            table3['aperture_sum_err'], atol=0.1)
        assert np.all(table1['aperture_sum_err'] < table3['aperture_sum_err'])


class TestCircular(BaseTestAperturePhotometry):
    def setup_class(self):
        self.data = np.ones((40, 40), dtype=float)
        position = (20.0, 20.0)
        r = 10.0
        self.aperture = CircularAperture(position, r)
        self.area = np.pi * r * r
        self.true_flux = self.area


class TestCircularArray(BaseTestAperturePhotometry):
    def setup_class(self):
        self.data = np.ones((40, 40), dtype=float)
        position = ((20.0, 20.0), (25.0, 25.0))
        r = 10.0
        self.aperture = CircularAperture(position, r)
        self.area = np.pi * r * r
        self.area = np.array((self.area,) * 2)
        self.true_flux = self.area


class TestCircularAnnulus(BaseTestAperturePhotometry):
    def setup_class(self):
        self.data = np.ones((40, 40), dtype=float)
        position = (20.0, 20.0)
        r_in = 8.0
        r_out = 10.0
        self.aperture = CircularAnnulus(position, r_in, r_out)
        self.area = np.pi * (r_out * r_out - r_in * r_in)
        self.true_flux = self.area


class TestCircularAnnulusArray(BaseTestAperturePhotometry):
    def setup_class(self):
        self.data = np.ones((40, 40), dtype=float)
        position = ((20.0, 20.0), (25.0, 25.0))
        r_in = 8.0
        r_out = 10.0
        self.aperture = CircularAnnulus(position, r_in, r_out)
        self.area = np.pi * (r_out * r_out - r_in * r_in)
        self.area = np.array((self.area,) * 2)
        self.true_flux = self.area


class TestElliptical(BaseTestAperturePhotometry):
    def setup_class(self):
        self.data = np.ones((40, 40), dtype=float)
        position = (20.0, 20.0)
        a = 10.0
        b = 5.0
        theta = -np.pi / 4.0
        self.aperture = EllipticalAperture(position, a, b, theta=theta)
        self.area = np.pi * a * b
        self.true_flux = self.area


class TestEllipticalAnnulus(BaseTestAperturePhotometry):
    def setup_class(self):
        self.data = np.ones((40, 40), dtype=float)
        position = (20.0, 20.0)
        a_in = 5.0
        a_out = 8.0
        b_out = 5.0
        theta = -np.pi / 4.0
        self.aperture = EllipticalAnnulus(position, a_in, a_out, b_out,
                                          theta=theta)
        self.area = (np.pi * (a_out * b_out)
                     - np.pi * (a_in * b_out * a_in / a_out))
        self.true_flux = self.area


class TestRectangularAperture(BaseTestAperturePhotometry):
    def setup_class(self):
        self.data = np.ones((40, 40), dtype=float)
        position = (20.0, 20.0)
        h = 5.0
        w = 8.0
        theta = np.pi / 4.0
        self.aperture = RectangularAperture(position, w, h, theta=theta)
        self.area = h * w
        self.true_flux = self.area


class TestRectangularAnnulus(BaseTestAperturePhotometry):
    def setup_class(self):
        self.data = np.ones((40, 40), dtype=float)
        position = (20.0, 20.0)
        h_out = 8.0
        w_in = 8.0
        w_out = 12.0
        h_in = w_in * h_out / w_out
        theta = np.pi / 8.0
        self.aperture = RectangularAnnulus(position, w_in, w_out, h_out,
                                           theta=theta)
        self.area = h_out * w_out - h_in * w_in
        self.true_flux = self.area


class TestMaskedSkipCircular(BaseTestAperturePhotometry):
    def setup_class(self):
        self.data = np.ones((40, 40), dtype=float)
        self.mask = np.zeros((40, 40), dtype=bool)
        self.mask[20, 20] = True
        position = (20.0, 20.0)
        r = 10.0
        self.aperture = CircularAperture(position, r)
        self.area = np.pi * r * r
        self.true_flux = self.area - 1


class BaseTestDifferentData:
    def test_basic_circular_aperture_photometry(self):
        aperture = CircularAperture(self.position, self.radius)
        table = aperture_photometry(self.data, aperture,
                                    method='exact')

        assert_allclose(table['aperture_sum'].value, self.true_flux)
        assert table['aperture_sum'].unit, self.fluxunit

        assert np.all(table['x_center'].value
                      == np.transpose(self.position)[0])
        assert np.all(table['y_center'].value
                      == np.transpose(self.position)[1])


class TestInputNDData(BaseTestDifferentData):
    def setup_class(self):
        data = np.ones((40, 40), dtype=float)
        self.data = NDData(data, unit=u.adu)
        self.radius = 3
        self.position = [(20, 20), (30, 30)]
        self.true_flux = np.pi * self.radius * self.radius
        self.fluxunit = u.adu


def test_input_wcs():
    data = make_4gaussians_image()
    wcs = make_wcs(data.shape)

    # Hard wired positions in make_4gaussian_image
    xypos = np.transpose(([160.0, 25.0, 150.0, 90.0],
                          [70.0, 40.0, 25.0, 60.0]))
    aper = CircularAperture(xypos, 3.0)
    aper3 = [CircularAperture((160.0, 70.0), r) for r in (3, 4, 5)]

    phot1 = aperture_photometry(data, aper[0], wcs=wcs)
    phot2 = aperture_photometry(data, aper, wcs=wcs)
    phot3 = aperture_photometry(data, aper3, wcs=wcs)

    assert 'sky_center' in phot1.colnames
    assert 'sky_center' in phot2.colnames
    assert 'sky_center' in phot3.colnames


def test_wcs_based_photometry():
    data = make_4gaussians_image()
    wcs = make_wcs(data.shape)

    # Hard wired positions in make_4gaussian_image
    pos_orig_pixel = u.Quantity(([160.0, 25.0, 150.0, 90.0],
                                 [70.0, 40.0, 25.0, 60.0]), unit=u.pixel)

    pos_skycoord = wcs.pixel_to_world(pos_orig_pixel[0], pos_orig_pixel[1])
    pos_skycoord_s = pos_skycoord[2]

    photometry_skycoord_circ = aperture_photometry(
        data, SkyCircularAperture(pos_skycoord, 3 * u.arcsec), wcs=wcs)
    photometry_skycoord_circ_2 = aperture_photometry(
        data, SkyCircularAperture(pos_skycoord, 2 * u.arcsec), wcs=wcs)
    photometry_skycoord_circ_s = aperture_photometry(
        data, SkyCircularAperture(pos_skycoord_s, 3 * u.arcsec), wcs=wcs)

    assert_allclose(photometry_skycoord_circ['aperture_sum'][2],
                    photometry_skycoord_circ_s['aperture_sum'])

    photometry_skycoord_circ_ann = aperture_photometry(
        data, SkyCircularAnnulus(pos_skycoord, 2 * u.arcsec, 3 * u.arcsec),
        wcs=wcs)
    photometry_skycoord_circ_ann_s = aperture_photometry(
        data, SkyCircularAnnulus(pos_skycoord_s, 2 * u.arcsec, 3 * u.arcsec),
        wcs=wcs)

    assert_allclose(photometry_skycoord_circ_ann['aperture_sum'][2],
                    photometry_skycoord_circ_ann_s['aperture_sum'])

    assert_allclose(photometry_skycoord_circ_ann['aperture_sum'],
                    photometry_skycoord_circ['aperture_sum']
                    - photometry_skycoord_circ_2['aperture_sum'])

    photometry_skycoord_ell = aperture_photometry(
        data, SkyEllipticalAperture(pos_skycoord, 3 * u.arcsec,
                                    3.0001 * u.arcsec, theta=45 * u.arcsec),
        wcs=wcs)
    photometry_skycoord_ell_2 = aperture_photometry(
        data, SkyEllipticalAperture(pos_skycoord, 2 * u.arcsec,
                                    2.0001 * u.arcsec, theta=45 * u.arcsec),
        wcs=wcs)
    photometry_skycoord_ell_s = aperture_photometry(
        data, SkyEllipticalAperture(pos_skycoord_s, 3 * u.arcsec,
                                    3.0001 * u.arcsec, theta=45 * u.arcsec),
        wcs=wcs)
    photometry_skycoord_ell_ann = aperture_photometry(
        data, SkyEllipticalAnnulus(pos_skycoord, 2 * u.arcsec, 3 * u.arcsec,
                                   3.0001 * u.arcsec, theta=45 * u.arcsec),
        wcs=wcs)
    photometry_skycoord_ell_ann_s = aperture_photometry(
        data, SkyEllipticalAnnulus(pos_skycoord_s, 2 * u.arcsec, 3 * u.arcsec,
                                   3.0001 * u.arcsec, theta=45 * u.arcsec),
        wcs=wcs)

    assert_allclose(photometry_skycoord_ell['aperture_sum'][2],
                    photometry_skycoord_ell_s['aperture_sum'])

    assert_allclose(photometry_skycoord_ell_ann['aperture_sum'][2],
                    photometry_skycoord_ell_ann_s['aperture_sum'])

    assert_allclose(photometry_skycoord_ell['aperture_sum'],
                    photometry_skycoord_circ['aperture_sum'], rtol=5e-3)

    assert_allclose(photometry_skycoord_ell_ann['aperture_sum'],
                    photometry_skycoord_ell['aperture_sum']
                    - photometry_skycoord_ell_2['aperture_sum'], rtol=1e-4)

    photometry_skycoord_rec = aperture_photometry(
        data, SkyRectangularAperture(pos_skycoord,
                                     6 * u.arcsec, 6 * u.arcsec,
                                     theta=0 * u.arcsec),
        method='subpixel', subpixels=20, wcs=wcs)
    photometry_skycoord_rec_4 = aperture_photometry(
        data, SkyRectangularAperture(pos_skycoord,
                                     4 * u.arcsec, 4 * u.arcsec,
                                     theta=0 * u.arcsec),
        method='subpixel', subpixels=20, wcs=wcs)
    photometry_skycoord_rec_s = aperture_photometry(
        data, SkyRectangularAperture(pos_skycoord_s,
                                     6 * u.arcsec, 6 * u.arcsec,
                                     theta=0 * u.arcsec),
        method='subpixel', subpixels=20, wcs=wcs)
    photometry_skycoord_rec_ann = aperture_photometry(
        data, SkyRectangularAnnulus(pos_skycoord, 4 * u.arcsec, 6 * u.arcsec,
                                    6 * u.arcsec, theta=0 * u.arcsec),
        method='subpixel', subpixels=20, wcs=wcs)
    photometry_skycoord_rec_ann_s = aperture_photometry(
        data, SkyRectangularAnnulus(pos_skycoord_s, 4 * u.arcsec,
                                    6 * u.arcsec, 6 * u.arcsec,
                                    theta=0 * u.arcsec),
        method='subpixel', subpixels=20, wcs=wcs)

    assert_allclose(photometry_skycoord_rec['aperture_sum'][2],
                    photometry_skycoord_rec_s['aperture_sum'])

    assert np.all(photometry_skycoord_rec['aperture_sum']
                  > photometry_skycoord_circ['aperture_sum'])

    assert_allclose(photometry_skycoord_rec_ann['aperture_sum'][2],
                    photometry_skycoord_rec_ann_s['aperture_sum'])

    assert_allclose(photometry_skycoord_rec_ann['aperture_sum'],
                    photometry_skycoord_rec['aperture_sum']
                    - photometry_skycoord_rec_4['aperture_sum'], rtol=1e-4)


def test_basic_circular_aperture_photometry_unit():
    radius = 3
    true_flux = np.pi * radius * radius
    aper = CircularAperture((12, 12), radius)

    data1 = np.ones((25, 25), dtype=float)
    table1 = aperture_photometry(data1, aper)
    assert_allclose(table1['aperture_sum'], true_flux)

    unit = u.adu
    data2 = u.Quantity(data1 * unit)
    table2 = aperture_photometry(data2, aper)
    assert_allclose(table2['aperture_sum'].value, true_flux)
    assert table2['aperture_sum'].unit == data2.unit == unit

    error1 = np.ones((25, 25))
    match = 'then they both must have the same units'
    with pytest.raises(ValueError, match=match):
        # data has unit, but error does not
        aperture_photometry(data2, aper, error=error1)

    error2 = u.Quantity(error1 * u.Jy)
    with pytest.raises(ValueError, match=match):
        # data and error have different units
        aperture_photometry(data2, aper, error=error2)


def test_aperture_photometry_with_error_units():
    """
    Test aperture_photometry when error has units (see #176).
    """
    data1 = np.ones((40, 40), dtype=float)
    data2 = u.Quantity(data1, unit=u.adu)
    error = u.Quantity(data1, unit=u.adu)
    radius = 3
    true_flux = np.pi * radius * radius
    unit = u.adu
    position = (20, 20)
    table1 = aperture_photometry(data2, CircularAperture(position, radius),
                                 error=error)
    assert_allclose(table1['aperture_sum'].value, true_flux)
    assert_allclose(table1['aperture_sum_err'].value, np.sqrt(true_flux))
    assert table1['aperture_sum'].unit == unit
    assert table1['aperture_sum_err'].unit == unit


def test_aperture_photometry_inputs_with_mask():
    """
    Test that aperture_photometry does not modify the input data or
    error array when a mask is input.
    """
    data = np.ones((5, 5))
    aperture = CircularAperture((2, 2), 2.0)
    mask = np.zeros_like(data, dtype=bool)
    data[2, 2] = 100.0  # bad pixel
    mask[2, 2] = True
    error = np.sqrt(data)
    data_in = data.copy()
    error_in = error.copy()
    t1 = aperture_photometry(data, aperture, error=error, mask=mask)
    assert_equal(data, data_in)
    assert_equal(error, error_in)
    assert_allclose(t1['aperture_sum'][0], 11.5663706144)
    t2 = aperture_photometry(data, aperture)
    assert_allclose(t2['aperture_sum'][0], 111.566370614)


TEST_ELLIPSE_EXACT_APERTURES = [(3.469906, 3.923861394, 3.0),
                                (0.3834415188257778, 0.3834415188257778, 0.3)]


@pytest.mark.parametrize(('x', 'y', 'r'), TEST_ELLIPSE_EXACT_APERTURES)
def test_ellipse_exact_grid(x, y, r):
    """
    Test elliptical exact aperture photometry on a grid of pixel
    positions.

    This is a regression test for the bug discovered in this issue:
    https://github.com/astropy/photutils/issues/198
    """
    data = np.ones((10, 10))

    aperture = EllipticalAperture((x, y), a=r, b=r, theta=0.0)
    t = aperture_photometry(data, aperture, method='exact')
    actual = t['aperture_sum'][0] / (np.pi * r**2)
    assert_allclose(actual, 1)


@pytest.mark.parametrize('value', [np.nan, np.inf])
def test_nan_inf_mask(value):
    """
    Test that nans and infs are properly masked [#267].
    """
    data = np.ones((9, 9))
    mask = np.zeros_like(data, dtype=bool)
    data[4, 4] = value
    mask[4, 4] = True
    radius = 2.0
    aper = CircularAperture((4, 4), radius)
    tbl = aperture_photometry(data, aper, mask=mask)
    desired = (np.pi * radius**2) - 1
    assert_allclose(tbl['aperture_sum'], desired)


def test_aperture_partial_overlap():
    data = np.ones((20, 20))
    error = np.ones((20, 20))
    xypos = [(10, 10), (0, 0), (0, 19), (19, 0), (19, 19)]
    r = 5.0
    aper = CircularAperture(xypos, r=r)
    tbl = aperture_photometry(data, aper, error=error)
    assert_allclose(tbl['aperture_sum'][0], np.pi * r**2)
    assert_array_less(tbl['aperture_sum'][1:], np.pi * r**2)

    unit = u.MJy / u.sr
    tbl = aperture_photometry(data * unit, aper, error=error * unit)
    assert_allclose(tbl['aperture_sum'][0].value, np.pi * r**2)
    assert_array_less(tbl['aperture_sum'][1:].value, np.pi * r**2)
    assert_array_less(tbl['aperture_sum_err'][1:].value, np.pi * r**2)
    assert tbl['aperture_sum'].unit == unit
    assert tbl['aperture_sum_err'].unit == unit


def test_pixel_aperture_repr():
    aper = CircularAperture((10, 20), r=3.0)
    assert '<CircularAperture(' in repr(aper)
    assert 'Aperture: CircularAperture' in str(aper)

    aper = CircularAnnulus((10, 20), r_in=3.0, r_out=5.0)
    assert '<CircularAnnulus(' in repr(aper)
    assert 'Aperture: CircularAnnulus' in str(aper)

    aper = EllipticalAperture((10, 20), a=5.0, b=3.0, theta=15.0)
    assert '<EllipticalAperture(' in repr(aper)
    assert 'Aperture: EllipticalAperture' in str(aper)

    aper = EllipticalAnnulus((10, 20), a_in=4.0, a_out=8.0, b_out=4.0,
                             theta=15.0)
    assert '<EllipticalAnnulus(' in repr(aper)
    assert 'Aperture: EllipticalAnnulus' in str(aper)

    aper = RectangularAperture((10, 20), w=5.0, h=3.0, theta=15.0)
    assert '<RectangularAperture(' in repr(aper)
    assert 'Aperture: RectangularAperture' in str(aper)

    aper = RectangularAnnulus((10, 20), w_in=4.0, w_out=8.0, h_out=4.0,
                              theta=15.0)
    assert '<RectangularAnnulus(' in repr(aper)
    assert 'Aperture: RectangularAnnulus' in str(aper)


def test_sky_aperture_repr():
    s = SkyCoord([1, 2], [3, 4], unit='deg')

    aper = SkyCircularAperture(s, r=3 * u.deg)
    a_repr = ('<SkyCircularAperture(<SkyCoord (ICRS): (ra, dec) in deg\n'
              '    [(1., 3.), (2., 4.)]>, r=3.0 deg)>')
    a_str = ('Aperture: SkyCircularAperture\npositions: <SkyCoord '
             '(ICRS): (ra, dec) in deg\n    [(1., 3.), (2., 4.)]>\n'
             'r: 3.0 deg')

    assert repr(aper) == a_repr
    assert str(aper) == a_str

    aper = SkyCircularAnnulus(s, r_in=3.0 * u.deg, r_out=5 * u.deg)
    a_repr = ('<SkyCircularAnnulus(<SkyCoord (ICRS): (ra, dec) in deg\n'
              '    [(1., 3.), (2., 4.)]>, r_in=3.0 deg, r_out=5.0 deg)>')
    a_str = ('Aperture: SkyCircularAnnulus\npositions: <SkyCoord '
             '(ICRS): (ra, dec) in deg\n    [(1., 3.), (2., 4.)]>\n'
             'r_in: 3.0 deg\nr_out: 5.0 deg')

    assert repr(aper) == a_repr
    assert str(aper) == a_str

    aper = SkyEllipticalAperture(s, a=3 * u.deg, b=5 * u.deg, theta=15 * u.deg)
    a_repr = ('<SkyEllipticalAperture(<SkyCoord (ICRS): (ra, dec) in '
              'deg\n    [(1., 3.), (2., 4.)]>, a=3.0 deg, b=5.0 deg, '
              'theta=15.0 deg)>')
    a_str = ('Aperture: SkyEllipticalAperture\npositions: <SkyCoord '
             '(ICRS): (ra, dec) in deg\n    [(1., 3.), (2., 4.)]>\n'
             'a: 3.0 deg\nb: 5.0 deg\ntheta: 15.0 deg')

    assert repr(aper) == a_repr
    assert str(aper) == a_str

    aper = SkyEllipticalAnnulus(s, a_in=3 * u.deg, a_out=5 * u.deg,
                                b_out=3 * u.deg, theta=15 * u.deg)
    a_repr = ('<SkyEllipticalAnnulus(<SkyCoord (ICRS): (ra, dec) in '
              'deg\n    [(1., 3.), (2., 4.)]>, a_in=3.0 deg, '
              'a_out=5.0 deg, b_in=1.8 deg, b_out=3.0 deg, '
              'theta=15.0 deg)>')
    a_str = ('Aperture: SkyEllipticalAnnulus\npositions: <SkyCoord '
             '(ICRS): (ra, dec) in deg\n    [(1., 3.), (2., 4.)]>\n'
             'a_in: 3.0 deg\na_out: 5.0 deg\nb_in: 1.8 deg\n'
             'b_out: 3.0 deg\ntheta: 15.0 deg')

    assert repr(aper) == a_repr
    assert str(aper) == a_str

    aper = SkyRectangularAperture(s, w=3 * u.deg, h=5 * u.deg,
                                  theta=15 * u.deg)
    a_repr = ('<SkyRectangularAperture(<SkyCoord (ICRS): (ra, dec) in '
              'deg\n    [(1., 3.), (2., 4.)]>, w=3.0 deg, h=5.0 deg'
              ', theta=15.0 deg)>')
    a_str = ('Aperture: SkyRectangularAperture\npositions: <SkyCoord '
             '(ICRS): (ra, dec) in deg\n    [(1., 3.), (2., 4.)]>\n'
             'w: 3.0 deg\nh: 5.0 deg\ntheta: 15.0 deg')

    assert repr(aper) == a_repr
    assert str(aper) == a_str

    aper = SkyRectangularAnnulus(s, w_in=5 * u.deg, w_out=10 * u.deg,
                                 h_out=6 * u.deg, theta=15 * u.deg)
    a_repr = ('<SkyRectangularAnnulus(<SkyCoord (ICRS): (ra, dec) in deg'
              '\n    [(1., 3.), (2., 4.)]>, w_in=5.0 deg, '
              'w_out=10.0 deg, h_in=3.0 deg, h_out=6.0 deg, '
              'theta=15.0 deg)>')
    a_str = ('Aperture: SkyRectangularAnnulus\npositions: <SkyCoord '
             '(ICRS): (ra, dec) in deg\n    [(1., 3.), (2., 4.)]>\n'
             'w_in: 5.0 deg\nw_out: 10.0 deg\nh_in: 3.0 deg\n'
             'h_out: 6.0 deg\ntheta: 15.0 deg')

    assert repr(aper) == a_repr
    assert str(aper) == a_str


def test_rectangular_bbox():
    # Test odd sizes
    width = 7
    height = 3
    a = RectangularAperture((50, 50), w=width, h=height, theta=0)
    assert a.bbox.shape == (height, width)

    a = RectangularAperture((50.5, 50.5), w=width, h=height, theta=0)
    assert a.bbox.shape == (height + 1, width + 1)

    a = RectangularAperture((50, 50), w=width, h=height,
                            theta=np.deg2rad(90.0))
    assert a.bbox.shape == (width, height)

    # Test even sizes
    width = 8
    height = 4
    a = RectangularAperture((50, 50), w=width, h=height, theta=0)
    assert a.bbox.shape == (height + 1, width + 1)

    a = RectangularAperture((50.5, 50.5), w=width, h=height, theta=0)
    assert a.bbox.shape == (height, width)

    a = RectangularAperture((50.5, 50.5), w=width, h=height,
                            theta=np.deg2rad(90.0))
    assert a.bbox.shape == (width, height)


def test_elliptical_bbox():
    # Integer axes
    a = 7
    b = 3
    ap = EllipticalAperture((50, 50), a=a, b=b, theta=0)
    assert ap.bbox.shape == (2 * b + 1, 2 * a + 1)

    ap = EllipticalAperture((50.5, 50.5), a=a, b=b, theta=0)
    assert ap.bbox.shape == (2 * b, 2 * a)

    ap = EllipticalAperture((50, 50), a=a, b=b, theta=np.deg2rad(90.0))
    assert ap.bbox.shape == (2 * a + 1, 2 * b + 1)

    # Fractional axes
    a = 7.5
    b = 4.5
    ap = EllipticalAperture((50, 50), a=a, b=b, theta=0)
    assert ap.bbox.shape == (2 * b, 2 * a)

    ap = EllipticalAperture((50.5, 50.5), a=a, b=b, theta=0)
    assert ap.bbox.shape == (2 * b + 1, 2 * a + 1)

    ap = EllipticalAperture((50, 50), a=a, b=b, theta=np.deg2rad(90.0))
    assert ap.bbox.shape == (2 * a, 2 * b)


@pytest.mark.skipif(not HAS_GWCS, reason='gwcs is required')
@pytest.mark.parametrize('wcs_type', ['wcs', 'gwcs'])
def test_to_sky_pixel(wcs_type):
    data = make_4gaussians_image()

    if wcs_type == 'wcs':
        wcs = make_wcs(data.shape)
    elif wcs_type == 'gwcs':
        wcs = make_gwcs(data.shape)

    ap = CircularAperture(((12.3, 15.7), (48.19, 98.14)), r=3.14)
    ap2 = ap.to_sky(wcs).to_pixel(wcs)
    assert_allclose(ap.positions, ap2.positions)
    assert_allclose(ap.r, ap2.r)

    ap = CircularAnnulus(((12.3, 15.7), (48.19, 98.14)), r_in=3.14,
                         r_out=5.32)
    ap2 = ap.to_sky(wcs).to_pixel(wcs)
    assert_allclose(ap.positions, ap2.positions)
    assert_allclose(ap.r_in, ap2.r_in)
    assert_allclose(ap.r_out, ap2.r_out)

    ap = EllipticalAperture(((12.3, 15.7), (48.19, 98.14)), a=3.14, b=5.32,
                            theta=np.deg2rad(103.0))
    ap2 = ap.to_sky(wcs).to_pixel(wcs)
    assert_allclose(ap.positions, ap2.positions)
    assert_allclose(ap.a, ap2.a)
    assert_allclose(ap.b, ap2.b)
    assert_allclose(ap.theta, ap2.theta)

    ap = EllipticalAnnulus(((12.3, 15.7), (48.19, 98.14)), a_in=3.14,
                           a_out=15.32, b_out=4.89,
                           theta=np.deg2rad(103.0))
    ap2 = ap.to_sky(wcs).to_pixel(wcs)
    assert_allclose(ap.positions, ap2.positions)
    assert_allclose(ap.a_in, ap2.a_in)
    assert_allclose(ap.a_out, ap2.a_out)
    assert_allclose(ap.b_out, ap2.b_out)
    assert_allclose(ap.theta, ap2.theta)

    ap = RectangularAperture(((12.3, 15.7), (48.19, 98.14)), w=3.14, h=5.32,
                             theta=np.deg2rad(103.0))
    ap2 = ap.to_sky(wcs).to_pixel(wcs)
    assert_allclose(ap.positions, ap2.positions)
    assert_allclose(ap.w, ap2.w)
    assert_allclose(ap.h, ap2.h)
    assert_allclose(ap.theta, ap2.theta)

    ap = RectangularAnnulus(((12.3, 15.7), (48.19, 98.14)), w_in=3.14,
                            w_out=15.32, h_out=4.89,
                            theta=np.deg2rad(103.0))
    ap2 = ap.to_sky(wcs).to_pixel(wcs)
    assert_allclose(ap.positions, ap2.positions)
    assert_allclose(ap.w_in, ap2.w_in)
    assert_allclose(ap.w_out, ap2.w_out)
    assert_allclose(ap.h_out, ap2.h_out)
    assert_allclose(ap.theta, ap2.theta)


def test_scalar_aperture():
    """
    Regression test to check that a length-1 aperture list appends a
    "_0" to the column names to be consistent with list inputs.
    """
    data = np.ones((20, 20), dtype=float)

    ap = CircularAperture((10, 10), r=3.0)
    colnames1 = aperture_photometry(data, ap, error=data).colnames
    assert (colnames1 == ['id', 'x_center', 'y_center', 'aperture_sum',
                          'aperture_sum_err'])

    colnames2 = aperture_photometry(data, [ap], error=data).colnames
    assert (colnames2 == ['id', 'x_center', 'y_center', 'aperture_sum_0',
                          'aperture_sum_err_0'])

    colnames3 = aperture_photometry(data, [ap, ap], error=data).colnames
    assert (colnames3 == ['id', 'x_center', 'y_center', 'aperture_sum_0',
                          'aperture_sum_err_0', 'aperture_sum_1',
                          'aperture_sum_err_1'])


def test_nan_in_bbox():
    """
    Regression test that non-finite data values outside the aperture
    mask but within the bounding box do not affect the photometry.
    """
    data1 = np.ones((101, 101))
    data2 = data1.copy()
    data1[33, 33] = np.nan
    data1[67, 67] = np.inf
    data1[33, 67] = -np.inf
    data1[22, 22] = np.nan
    data1[22, 23] = np.inf
    error = data1.copy()

    aper1 = CircularAperture((50, 50), r=20.0)
    aper2 = CircularAperture((5, 5), r=20.0)

    tbl1 = aperture_photometry(data1, aper1, error=error)
    tbl2 = aperture_photometry(data2, aper1, error=error)
    assert_allclose(tbl1['aperture_sum'], tbl2['aperture_sum'])
    assert_allclose(tbl1['aperture_sum_err'], tbl2['aperture_sum_err'])

    tbl3 = aperture_photometry(data1, aper2, error=error)
    tbl4 = aperture_photometry(data2, aper2, error=error)
    assert_allclose(tbl3['aperture_sum'], tbl4['aperture_sum'])
    assert_allclose(tbl3['aperture_sum_err'], tbl4['aperture_sum_err'])


def test_scalar_skycoord():
    """
    Regression test to check that scalar SkyCoords are added to the
    table as a length-1 SkyCoord array.
    """
    data = make_4gaussians_image()
    wcs = make_wcs(data.shape)
    skycoord = wcs.pixel_to_world(90, 60)
    aper = SkyCircularAperture(skycoord, r=0.1 * u.arcsec)
    tbl = aperture_photometry(data, aper, wcs=wcs)
    assert isinstance(tbl['sky_center'], SkyCoord)


@pytest.mark.parametrize('units', [False, True])
def test_nddata_input(units):
    data = np.arange(400).reshape((20, 20))
    error = np.sqrt(data)
    mask = np.zeros((20, 20), dtype=bool)
    mask[8:13, 8:13] = True
    if units:
        unit = u.Jy
        data = data * unit
        error = error * unit
    else:
        unit = None

    wcs = make_wcs(data.shape)
    skycoord = wcs.pixel_to_world(10, 10)
    aper = SkyCircularAperture(skycoord, r=0.7 * u.arcsec)

    tbl1 = aperture_photometry(data, aper, error=error, mask=mask, wcs=wcs)

    uncertainty = StdDevUncertainty(error)
    nddata = NDData(data, uncertainty=uncertainty, mask=mask, wcs=wcs,
                    unit=unit)
    tbl2 = aperture_photometry(nddata, aper)

    for column in tbl1.columns:
        if column == 'sky_center':  # cannot test SkyCoord equality
            continue
        assert_allclose(tbl1[column], tbl2[column])

    match = 'keyword is ignored. Its value is obtained from the input'
    with pytest.warns(AstropyUserWarning, match=match):
        aperture_photometry(nddata, aper, error=error)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
class BaseTestRegionPhotometry:
    def test_region_matches_aperture(self):
        data = np.ones((40, 40), dtype=float)
        error = np.ones(data.shape, dtype=float)
        region_tables = [
            aperture_photometry(data, self.region, method='center',
                                error=error),
            aperture_photometry(data, self.region,
                                method='subpixel', subpixels=12,
                                error=error),
            aperture_photometry(data, self.region, method='exact',
                                error=error),
        ]
        aperture_tables = [
            aperture_photometry(data, self.aperture, method='center',
                                error=error),
            aperture_photometry(data, self.aperture,
                                method='subpixel', subpixels=12,
                                error=error),
            aperture_photometry(data, self.aperture, method='exact',
                                error=error),
        ]

        for reg_table, ap_table in zip(region_tables, aperture_tables,
                                       strict=True):
            assert_allclose(reg_table['aperture_sum'],
                            ap_table['aperture_sum'])

        if isinstance(self.aperture, (RectangularAperture,
                                      RectangularAnnulus)):
            for reg_table, ap_table in zip(region_tables, aperture_tables,
                                           strict=True):
                assert_allclose(reg_table['aperture_sum_err'],
                                ap_table['aperture_sum_err'])


def test_invalid_inputs():
    data = np.ones((11, 11))
    aper = CircularAperture((5, 5), r=3)
    wcs = make_wcs(data.shape)
    sky_aper = aper.to_sky(wcs=wcs)

    match = 'A WCS transform must be defined'
    with pytest.raises(ValueError, match=match):
        aperture_photometry(data, sky_aper)

    aper2 = CircularAperture((7, 7), r=3)
    sky_aper2 = aper2.to_sky(wcs=wcs)
    apers = [aper, aper2]
    sky_apers = [sky_aper, sky_aper2]

    match = 'Input apertures must all have identical positions'
    with pytest.raises(ValueError, match=match):
        aperture_photometry(data, apers)
    with pytest.raises(ValueError, match=match):
        aperture_photometry(data, sky_apers, wcs=wcs)

    data = np.ones((11, 11))
    aper = CircularAperture((5, 5), r=3)
    match = 'subpixels must be a strictly positive integer'
    with pytest.raises(ValueError, match=match):
        aperture_photometry(data, aper, method='subpixel', subpixels=0)
    with pytest.raises(ValueError, match=match):
        aperture_photometry(data, aper, method='subpixel', subpixels=-1)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
class TestCircleRegionPhotometry(BaseTestRegionPhotometry):
    def setup_class(self):
        from regions import CirclePixelRegion, PixCoord
        position = (20.0, 20.0)
        r = 10.0
        self.region = CirclePixelRegion(PixCoord(*position), r)
        self.aperture = CircularAperture(position, r)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
class TestCircleAnnulusRegionPhotometry(BaseTestRegionPhotometry):
    def setup_class(self):
        from regions import CircleAnnulusPixelRegion, PixCoord
        position = (20.0, 20.0)
        r_in = 8.0
        r_out = 10.0
        self.region = CircleAnnulusPixelRegion(PixCoord(*position), r_in,
                                               r_out)
        self.aperture = CircularAnnulus(position, r_in, r_out)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
class TestEllipseRegionPhotometry(BaseTestRegionPhotometry):
    def setup_class(self):
        from regions import EllipsePixelRegion, PixCoord
        position = (20.0, 20.0)
        a = 10.0
        b = 5.0
        theta = (-np.pi / 4.0) * u.rad
        self.region = EllipsePixelRegion(PixCoord(*position), a * 2, b * 2,
                                         theta)
        self.aperture = EllipticalAperture(position, a, b, theta=theta)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
class TestEllipseAnnulusRegionPhotometry(BaseTestRegionPhotometry):
    def setup_class(self):
        from regions import EllipseAnnulusPixelRegion, PixCoord
        position = (20.0, 20.0)
        a_in = 5.0
        a_out = 8.0
        b_in = 3.0
        b_out = 5.0
        theta = (-np.pi / 4.0) * u.rad
        self.region = EllipseAnnulusPixelRegion(PixCoord(*position),
                                                a_in * 2, a_out * 2,
                                                b_in * 2, b_out * 2,
                                                theta)
        self.aperture = EllipticalAnnulus(position,
                                          a_in, a_out,
                                          b_out, b_in=b_in,
                                          theta=theta)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
class TestRectangleRegionPhotometry(BaseTestRegionPhotometry):
    def setup_class(self):
        from regions import PixCoord, RectanglePixelRegion
        position = (20.0, 20.0)
        h = 5.0
        w = 8.0
        theta = (np.pi / 4.0) * u.rad
        self.region = RectanglePixelRegion(PixCoord(*position), w, h, theta)
        self.aperture = RectangularAperture(position, w, h, theta=theta)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
class TestRectangleAnnulusRegionPhotometry(BaseTestRegionPhotometry):
    def setup_class(self):
        from regions import PixCoord, RectangleAnnulusPixelRegion
        position = (20.0, 20.0)
        h_out = 8.0
        w_in = 8.0
        w_out = 12.0
        h_in = w_in * h_out / w_out
        theta = (np.pi / 8.0) * u.rad
        self.region = RectangleAnnulusPixelRegion(PixCoord(*position),
                                                  w_in, w_out, h_in, h_out,
                                                  theta)
        self.aperture = RectangularAnnulus(position, w_in, w_out,
                                           h_out, h_in=h_in, theta=theta)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_unsupported_region_input():
    from regions import PixCoord, PolygonPixelRegion

    region = PolygonPixelRegion(vertices=PixCoord(x=[1, 2, 3], y=[1, 1, 2]))
    data = np.ones((10, 10))
    match = r'Cannot convert .* to an Aperture object'
    with pytest.raises(TypeError, match=match):
        aperture_photometry(data, region)


def test_aperture_metadata():
    x = [10, 20, 3]
    y = [3, 5, 10]
    xypos = list(zip(x, y, strict=False))
    a1 = CircularAperture(xypos, r=3)
    a2 = CircularAperture(xypos, r=4)
    a3 = CircularAnnulus(xypos, 5, 10)
    a4 = EllipticalAperture(xypos, 10, 5, theta=10 * u.deg)
    a5 = EllipticalAnnulus(xypos, a_in=5, a_out=10, b_in=3, b_out=5,
                           theta=20 * u.deg)
    a6 = RectangularAperture(xypos, 10, 5, theta=30 * u.deg)
    a7 = RectangularAnnulus(xypos, w_in=5, w_out=10, h_in=3, h_out=5,
                            theta=40 * u.deg)
    apers = (a1, a2, a3, a4, a5, a6, a7)
    data = np.ones((50, 50))
    tbl = aperture_photometry(data, apers)

    for i, aper in enumerate(apers):
        assert tbl.meta[f'aperture{i}'] == aper.__class__.__name__
        params = aper._params
        for param in params:
            if param != 'positions':
                assert tbl.meta[f'aperture{i}_{param}'] == getattr(aper, param)

    wcs = make_wcs(data.shape)
    skycoord = wcs.pixel_to_world(10, 10)
    unit = u.arcsec
    saper = SkyEllipticalAnnulus(skycoord, a_in=0.1 * unit, a_out=0.2 * unit,
                                 b_in=0.05 * unit, b_out=0.1 * unit,
                                 theta=10 * u.deg)
    tbl = aperture_photometry(data, saper, wcs=wcs)
    assert tbl.meta['aperture'] == saper.__class__.__name__
    assert tbl.meta['aperture_a_in'] == saper.a_in
    assert tbl.meta['aperture_a_out'] == saper.a_out
    assert tbl.meta['aperture_b_out'] == saper.b_out
    assert tbl.meta['aperture_theta'] == saper.theta



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for making simulated example images for documentation examples and
tests.
"""

import pathlib

import numpy as np
from astropy.modeling.models import Gaussian2D
from astropy.table import QTable
from astropy.utils.data import get_pkg_data_path

from photutils.datasets import make_model_image
from photutils.utils._deprecation import deprecated_positional_kwargs

__all__ = ['make_4gaussians_image', 'make_100gaussians_image']

_DATASETS_DATA_DIR = pathlib.Path(get_pkg_data_path('datasets', 'data',
                                                    package='photutils'))


def _make_gaussians_image(filename, shape, noise, noise_stddev,
                          **kwargs):
    """
    Make an image containing 2D Gaussians from a parameter file.

    Parameters
    ----------
    filename : str
        The name of the ECSV file containing the Gaussian parameters.

    shape : 2-tuple of int
        The shape of the output image.

    noise : bool
        Whether to include noise in the output image.

    noise_stddev : float
        The standard deviation of the Gaussian noise.

    **kwargs
        Additional keyword arguments passed to `make_model_image`.

    Returns
    -------
    image : 2D `~numpy.ndarray`
        Image containing 2D Gaussian sources.
    """
    model = Gaussian2D()
    params = QTable.read(_DATASETS_DATA_DIR / filename,
                         format='ascii.ecsv')
    data = make_model_image(shape, model, params, x_name='x_mean',
                            y_name='y_mean', **kwargs)
    data += 5.0  # background

    if noise:
        rng = np.random.default_rng(seed=0)
        data += rng.normal(loc=0.0, scale=noise_stddev, size=shape)

    return data


@deprecated_positional_kwargs(since='3.0', until='4.0')
def make_4gaussians_image(noise=True):
    """
    Make an example image containing four 2D Gaussians plus a constant
    background.

    The background has a mean of 5.

    If ``noise`` is `True`, then Gaussian noise with a mean of 0 and a
    standard deviation of 5 is added to the output image.

    Parameters
    ----------
    noise : bool, optional
        Whether to include noise in the output image (default is
        `True`).

    Returns
    -------
    image : 2D `~numpy.ndarray`
        Image containing four 2D Gaussian sources.

    See Also
    --------
    make_100gaussians_image

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from astropy.visualization import simple_norm
        from photutils.datasets import make_4gaussians_image

        image = make_4gaussians_image()
        fig, ax = plt.subplots()
        norm = simple_norm(image, 'sqrt', percent=99.5)
        ax.imshow(image, norm=norm, origin='lower')
    """
    return _make_gaussians_image('4gaussians_params.ecsv', (100, 200),
                                 noise, noise_stddev=5.0)


@deprecated_positional_kwargs(since='3.0', until='4.0')
def make_100gaussians_image(noise=True):
    """
    Make an example image containing 100 2D Gaussians plus a constant
    background.

    The background has a mean of 5.

    If ``noise`` is `True`, then Gaussian noise with a mean of 0 and a
    standard deviation of 2 is added to the output image.

    Parameters
    ----------
    noise : bool, optional
        Whether to include noise in the output image (default is
        `True`).

    Returns
    -------
    image : 2D `~numpy.ndarray`
        Image containing 100 2D Gaussian sources.

    See Also
    --------
    make_4gaussians_image

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from astropy.visualization import simple_norm
        from photutils.datasets import make_100gaussians_image

        image = make_100gaussians_image()
        fig, ax = plt.subplots()
        norm = simple_norm(image, 'sqrt', percent=99.5)
        ax.imshow(image, norm=norm, origin='lower')
    """
    return _make_gaussians_image('100gaussians_params.ecsv', (300, 500),
                                 noise, noise_stddev=2.0, bbox_factor=6.0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for progress bars.
"""

# pylint: disable-next=E0611
from photutils.utils._optional_deps import HAS_TQDM


# Allow iterable to be passed positionally
def add_progress_bar(iterable=None, *, desc=None, total=None, text=False):
    """
    Add a progress bar for an iterable.

    Parameters
    ----------
    iterable : iterable, optional
        The iterable for which to add a progress bar. Set to `None` to
        manually manage the progress bar updates.

    desc : str, optional
        The prefix string for the progress bar.

    total : int, optional
        The number of expected iterations. If unspecified, len(iterable)
        is used if possible.

    text : bool, optional
        Whether to always use a text-based progress bar.

    Returns
    -------
    result : tqdm iterable, iterable, or `None`
        When tqdm is installed, a tqdm progress bar (ipywidgets-based in
        a notebook if available, otherwise text-based). When tqdm is not
        installed, the original iterable is returned unchanged (which
        may be `None`).

    Notes
    -----
    When ``iterable=None`` and ``total`` is provided, the returned tqdm
    object is a manually-managed progress bar. The caller is responsible
    for advancing it by calling ``.update()`` and closing it when
    finished. This mode is used when the caller drives the loop directly
    (e.g., with a ``while`` loop) rather than iterating over a sequence.
    When tqdm is not installed, `None` is returned in this case and the
    caller must guard against it.
    """
    if HAS_TQDM:
        if text:
            from tqdm import tqdm
        else:
            try:
                # pylint: disable-next=W0611
                from ipywidgets import FloatProgress  # noqa: F401
                from tqdm.auto import tqdm
            except ImportError:
                from tqdm import tqdm

        iterable = tqdm(iterable=iterable, desc=desc,
                        total=total)

    return iterable


# Define tqdm as a dummy class if it is not available.
# This is needed to use tqdm as a context manager with multiprocessing.
try:
    from tqdm.auto import tqdm
except ImportError:
    class tqdm:  # noqa: N801
        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self):
            return self

        def __exit__(self, *exc):
            pass

        def update(self, *args, **kwargs):
            pass

        def set_postfix_str(self, *args, **kwargs):
            pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the parameters module.
"""


import numpy as np
import pytest
from astropy.stats import SigmaClip
from numpy.testing import assert_equal

from photutils.utils._parameters import (SigmaClipSentinelDefault, as_pair,
                                         create_default_sigmaclip)


class TestAsPairBasic:
    """
    Tests for as_pair scalar/tuple broadcasting and basic validation.
    """

    def test_scalar_broadcast(self):
        assert_equal(as_pair('p', 4), (4, 4))

    def test_tuple_passthrough(self):
        assert_equal(as_pair('p', (3, 4)), (3, 4))

    def test_scalar_zero(self):
        assert_equal(as_pair('p', 0), (0, 0))

    def test_too_many_elements(self):
        match = 'must have 1 or 2 elements'
        with pytest.raises(ValueError, match=match):
            as_pair('p', (1, 2, 3))

    def test_2d_input(self):
        match = 'must be 1D'
        with pytest.raises(ValueError, match=match):
            as_pair('p', np.array([[1, 2]]))

    def test_non_integer_dtype(self):
        match = 'must have integer values'
        with pytest.raises(ValueError, match=match):
            as_pair('p', 1.5)

    @pytest.mark.parametrize('value', [(1, np.nan), (1, np.inf)])
    def test_non_finite(self, value):
        match = 'must be a finite value'
        with pytest.raises(ValueError, match=match):
            as_pair('p', value)


class TestAsPairCheckOdd:
    """
    Tests for the check_odd parameter.
    """

    def test_odd_tuple(self):
        assert_equal(as_pair('p', (3, 5), check_odd=True), (3, 5))

    def test_odd_scalar(self):
        assert_equal(as_pair('p', 3, check_odd=True), (3, 3))

    @pytest.mark.parametrize('value', [(3, 4), 4])
    def test_even_raises(self, value):
        match = 'must have an odd value for both axes'
        with pytest.raises(ValueError, match=match):
            as_pair('p', value, check_odd=True)


class TestAsPairLowerBound:
    """
    Tests for lower_bound validation.
    """

    def test_exclusive_lower_bound(self):
        match = 'must be > 0'
        with pytest.raises(ValueError, match=match):
            as_pair('p', 0, lower_bound=(0, 0))

    def test_inclusive_lower_bound(self):
        result = as_pair('p', 0, lower_bound=(0, 1))
        assert_equal(result, (0, 0))

    def test_inclusive_lower_bound_violation(self):
        match = r'must be >= 1'
        with pytest.raises(ValueError, match=match):
            as_pair('p', 0, lower_bound=(1, 1))

    def test_lower_bound_wrong_length(self):
        match = 'lower_bound must contain only 2 elements'
        with pytest.raises(ValueError, match=match):
            as_pair('p', 1, lower_bound=(0,))


class TestAsPairUpperBound:
    """
    Tests for upper_bound validation and clamping.
    """

    def test_upper_bound_clipping(self):
        result = as_pair('p', (10, 20), upper_bound=(5, 15))
        assert_equal(result, (5, 15))

    def test_upper_bound_no_clipping(self):
        result = as_pair('p', (3, 4), upper_bound=(10, 10))
        assert_equal(result, (3, 4))

    def test_upper_bound_wrong_length(self):
        match = 'upper_bound must contain only 2 elements'
        with pytest.raises(ValueError, match=match):
            as_pair('p', (3, 4), upper_bound=(5,))


class TestSigmaClipSentinelDefault:
    """
    Tests for SigmaClipSentinelDefault.
    """

    def test_repr(self):
        sentinel = SigmaClipSentinelDefault(sigma=3.0, maxiters=10)
        result = repr(sentinel)
        assert 'SigmaClip' in result
        assert '3.0' in result
        assert '10' in result

    def test_custom_params(self):
        sentinel = SigmaClipSentinelDefault(sigma=2.0, maxiters=5)
        assert sentinel.sigma == 2.0
        assert sentinel.maxiters == 5


class TestCreateDefaultSigmaClip:
    """
    Tests for create_default_sigmaclip.
    """

    def test_defaults(self):
        sc = create_default_sigmaclip()
        assert isinstance(sc, SigmaClip)
        assert sc.sigma == 3.0
        assert sc.maxiters == 10

    def test_custom(self):
        sc = create_default_sigmaclip(sigma=2.5, maxiters=5)
        assert isinstance(sc, SigmaClip)
        assert sc.sigma == 2.5
        assert sc.maxiters == 5



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for deprecation warnings when optional arguments are passed
positionally.
"""

import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.profiles import CurveOfGrowth, RadialProfile
from photutils.utils._optional_deps import HAS_MATPLOTLIB


@pytest.fixture
def profile_data():
    """
    Create a simple radial profile for testing.
    """
    gmodel = Gaussian2D(42.1, 50, 50, 4.7, 4.7, 0)
    yy, xx = np.mgrid[0:100, 0:100]
    data = gmodel(xx, yy)
    xycen = (50, 50)
    return data, xycen


class TestProfileBaseNormalizePositionalKwargs:
    """
    Test that ProfileBase.normalize warns for positional optional args.
    """

    def test_positional_warns(self, profile_data):
        data, xycen = profile_data
        radii = np.arange(1, 20)
        cog = CurveOfGrowth(data, xycen, radii)
        match = 'normalize'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            cog.normalize('max')

    def test_keyword_no_warning(self, profile_data):
        data, xycen = profile_data
        radii = np.arange(1, 20)
        cog = CurveOfGrowth(data, xycen, radii)
        cog.normalize(method='max')


class TestProfileBasePlotPositionalKwargs:
    """
    Test that ProfileBase.plot warns for positional optional args.
    """

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_positional_warns(self, profile_data):
        data, xycen = profile_data
        edge_radii = np.arange(20)
        rp = RadialProfile(data, xycen, edge_radii)
        match = 'plot'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            rp.plot(None)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_keyword_no_warning(self, profile_data):
        data, xycen = profile_data
        edge_radii = np.arange(20)
        rp = RadialProfile(data, xycen, edge_radii)
        rp.plot(ax=None)


class TestProfileBasePlotErrorPositionalKwargs:
    """
    Test that ProfileBase.plot_error warns for positional optional args.
    """

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_positional_warns(self, profile_data):
        data, xycen = profile_data
        error = np.ones_like(data)
        edge_radii = np.arange(20)
        rp = RadialProfile(data, xycen, edge_radii, error=error)
        match = 'plot_error'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            rp.plot_error(None)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_keyword_no_warning(self, profile_data):
        data, xycen = profile_data
        error = np.ones_like(data)
        edge_radii = np.arange(20)
        rp = RadialProfile(data, xycen, edge_radii, error=error)
        rp.plot_error(ax=None)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Utility functions for the psf_matching subpackage.
"""


import numpy as np
from scipy.fft import fft2, fftshift, ifftshift
from scipy.ndimage import zoom

__all__ = ['resize_psf']


def _validate_kernel_inputs(source_psf, target_psf, window):
    """
    Validate and prepare common inputs for kernel-making functions.

    Parameters
    ----------
    source_psf : array-like
        The source PSF array.

    target_psf : array-like
        The target PSF array.

    window : callable or None
        The window function or None.

    Returns
    -------
    source_psf : `~numpy.ndarray`
        The validated and normalized source PSF as a float array.

    target_psf : `~numpy.ndarray`
        The validated and normalized target PSF as a float array.

    Raises
    ------
    ValueError
        If the PSFs are not 2D arrays, have even dimensions, do not have
        the same shape, or contain NaN or Inf values.

    TypeError
        If the input ``window`` is not callable.
    """
    # Copy as float so in-place normalization doesn't modify inputs
    source_psf = np.array(source_psf, dtype=float)
    target_psf = np.array(target_psf, dtype=float)

    _validate_psf(source_psf, 'source_psf')
    _validate_psf(target_psf, 'target_psf')

    if source_psf.shape != target_psf.shape:
        msg = ('source_psf and target_psf must have the same shape '
               '(i.e., registered with the same pixel scale).')
        raise ValueError(msg)

    if window is not None and not callable(window):
        msg = 'window must be a callable.'
        raise TypeError(msg)

    # Ensure input PSFs are normalized
    source_psf /= source_psf.sum()
    target_psf /= target_psf.sum()

    return source_psf, target_psf


def _validate_psf(psf, name):
    """
    Validate that a PSF is 2D with odd dimensions.

    Parameters
    ----------
    psf : `~numpy.ndarray`
        The PSF array to validate.

    name : str
        The parameter name used in error messages.

    Raises
    ------
    ValueError
        If the PSF is not 2D, has even dimensions, or contains NaN or
        Inf values.
    """
    if psf.ndim != 2:
        msg = f'{name} must be a 2D array.'
        raise ValueError(msg)

    if psf.shape[0] % 2 == 0 or psf.shape[1] % 2 == 0:
        msg = (f'{name} must have odd dimensions, got '
               f'shape {psf.shape}.')
        raise ValueError(msg)

    if not np.all(np.isfinite(psf)):
        msg = f'{name} contains NaN or Inf values.'
        raise ValueError(msg)

    if np.sum(psf) == 0:
        msg = f'{name} must have a non-zero sum; it cannot be normalized.'
        raise ValueError(msg)


def _validate_window_array(window_array, expected_shape):
    """
    Validate window function output.

    Parameters
    ----------
    window_array : any
        The array returned by the window function.

    expected_shape : tuple
        The expected shape of the window array.

    Raises
    ------
    ValueError
        If the window array is not a 2D array, has the wrong shape,
        or contains values outside the range [0, 1].
    """
    if not isinstance(window_array, np.ndarray) or window_array.ndim != 2:
        msg = ('window function must return a 2D array, got '
               f'{type(window_array).__name__} with '
               f'ndim={getattr(window_array, "ndim", "undefined")}.')
        raise ValueError(msg)

    if window_array.shape != expected_shape:
        msg = (f'window function must return an array with shape '
               f'{expected_shape}, got {window_array.shape}.')
        raise ValueError(msg)

    if np.any(np.logical_or(window_array < 0, window_array > 1)):
        msg = ('window function values must be in the range [0, 1], '
               f'got range [{np.min(window_array)}, '
               f'{np.max(window_array)}].')
        raise ValueError(msg)


def _convert_psf_to_otf(psf, shape):
    """
    Convert a point-spread function to an optical transfer function.

    This computes the FFT of a PSF array after centering it in a
    zero-padded array of the output shape and applying `ifftshift` to
    move the PSF center to position [0, 0].

    The PSF is first placed at the center of the zero-padded array,
    ensuring its center aligns with the array's center. The zero-padding
    is needed when the input kernel (e.g., a 3x3 Laplacian) is smaller
    than the target shape, so that the resulting OTF has the correct
    size for element-wise operations with other same-shaped OTFs.

    The `ifftshift` operation then moves the PSF center from the array
    center to position [0, 0], which is the standard convention for
    computing OTFs via FFT. This ensures correct complex phase in
    the resulting OTF for general use. Note that when only the power
    spectrum (|OTF|^2) is needed, the shift has no effect because it
    only changes the phase.

    Parameters
    ----------
    psf : 2D `~numpy.ndarray`
        The PSF array. The PSF must have odd dimensions and be centered
        on the central pixel. The PSF shape must be smaller than or
        equal to the target shape in both dimensions.

    shape : tuple of int
        The desired output shape.

    Returns
    -------
    otf : 2D `~numpy.ndarray`
        The optical transfer function (complex array).
    """
    if np.all(psf == 0):
        return np.zeros(shape, dtype=complex)

    if psf.ndim != 2:
        msg = 'psf must be a 2D array.'
        raise ValueError(msg)

    if psf.shape[0] % 2 == 0 or psf.shape[1] % 2 == 0:
        msg = f'psf must have odd dimensions, got shape {psf.shape}.'
        raise ValueError(msg)

    inshape = psf.shape

    if any(i > s for i, s in zip(inshape, shape, strict=True)):
        msg = (f'The PSF shape {inshape} is larger than the target '
               f'shape {shape} in at least one dimension.')
        raise ValueError(msg)

    # Zero-pad to the output shape with PSF centered in the array
    padded = np.zeros(shape, dtype=psf.dtype)

    # Calculate where to place PSF so its center aligns with padded
    # array center
    center = tuple(s // 2 for s in shape)
    psf_center = tuple(s // 2 for s in inshape)
    start = tuple(c - pc for c, pc in zip(center, psf_center, strict=True))
    padded[start[0]:start[0] + inshape[0],
           start[1]:start[1] + inshape[1]] = psf

    # Shift the centered PSF so its center moves to [0, 0]
    padded = ifftshift(padded)

    return fft2(padded)


def _apply_window_to_fourier(fourier_array, window, shape):
    """
    Apply a centered window function to a Fourier-domain array.

    The window function is assumed to be defined with the DC component
    at the center of the array. Since Fourier arrays use the standard
    FFT layout with the DC component at the corner, this function shifts
    the array to the center, applies the window, and shifts it back.

    Parameters
    ----------
    fourier_array : 2D `~numpy.ndarray`
        A complex Fourier-domain array with the DC component at the
        corner.

    window : callable
        The window function. Must accept a single ``shape`` tuple and
        return a 2D array with values in [0, 1].

    shape : tuple of int
        The shape passed to the window function and the expected shape
        of the window output.

    Returns
    -------
    result : 2D `~numpy.ndarray`
        The windowed Fourier-domain array, still in standard FFT layout
        (DC at the corner).
    """
    window_array = window(shape)
    _validate_window_array(window_array, shape)
    fourier_array = fftshift(fourier_array)
    fourier_array *= window_array
    return ifftshift(fourier_array)


def resize_psf(psf, input_pixel_scale, output_pixel_scale, *, order=3):
    """
    Resize a PSF using spline interpolation of the requested order.

    The total flux of the PSF is conserved during the resizing.

    Parameters
    ----------
    psf : 2D `~numpy.ndarray`
        The 2D data array of the PSF. The PSF must have odd dimensions.
        It is assumed to be centered on the central pixel.

    input_pixel_scale : float
        The pixel scale of the input ``psf``. The units must match
        ``output_pixel_scale``.

    output_pixel_scale : float
        The pixel scale of the output ``psf``. The units must match
        ``input_pixel_scale``.

    order : int, optional
        The order of the spline interpolation (0-5). The default is 3.

    Returns
    -------
    result : 2D `~numpy.ndarray`
        The resampled/interpolated 2D data array. The output always
        has odd dimensions. The natural resampled size is computed
        by taking the ceiling of ``input_size * (input_pixel_scale
        / output_pixel_scale)`` for each axis, then adding 1 to
        any axis whose size is even. This guarantees the output is
        centered and usable for PSF matching. When the output size is
        adjusted, the effective pixel scale will be slightly smaller
        than ``output_pixel_scale``; the exact value per axis is
        ``input_pixel_scale * input_size / output_size``.

    Raises
    ------
    ValueError
        If ``psf`` is not a 2D array, has even dimensions, is not
        centered, or if the pixel scales are not positive.
    """
    psf = np.asarray(psf, dtype=float)

    if input_pixel_scale <= 0 or output_pixel_scale <= 0:
        msg = ('input_pixel_scale and output_pixel_scale must be '
               'positive.')
        raise ValueError(msg)

    _validate_psf(psf, 'psf')

    ratio = input_pixel_scale / output_pixel_scale

    # Compute target shape using ceiling (never discard pixels), then
    # add 1 to any even dimension to guarantee an odd output, which is
    # required for PSF matching.
    in_shape = np.array(psf.shape)
    out_shape = np.maximum(1, np.ceil(in_shape * ratio).astype(int))
    out_shape += out_shape % 2 == 0

    # Per-axis zoom factors for the forced-odd target shape
    zoom_factors = out_shape / in_shape

    # Normalize the PSF to conserve total flux after resizing.
    psf_sum = psf.sum()
    result = zoom(psf, zoom_factors, order=order)
    return result * (psf_sum / result.sum())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the stats module.
"""

import sys
from unittest.mock import patch

import astropy.units as u
import numpy as np
import pytest
from astropy.nddata import NDData, StdDevUncertainty
from astropy.stats import SigmaClip
from astropy.utils.exceptions import (AstropyDeprecationWarning,
                                      AstropyUserWarning)
from numpy.testing import assert_allclose, assert_equal

from photutils.aperture.circle import CircularAnnulus, CircularAperture
from photutils.aperture.ellipse import (EllipticalAnnulus, EllipticalAperture,
                                        SkyEllipticalAnnulus)
from photutils.aperture.rectangle import (RectangularAnnulus,
                                          RectangularAperture)
from photutils.aperture.stats import ApertureStats
from photutils.datasets import make_100gaussians_image, make_wcs
from photutils.utils._optional_deps import HAS_REGIONS


class TestApertureStats:
    data = make_100gaussians_image()
    error = np.sqrt(np.abs(data))
    wcs = make_wcs(data.shape)
    positions = ((145.1, 168.3), (84.7, 224.1), (48.3, 200.3))
    aperture = CircularAperture(positions, r=5)

    sigclip = SigmaClip(sigma=3.0, maxiters=10)
    apstats1 = ApertureStats(data, aperture, error=error, wcs=wcs,
                             sigma_clip=None)
    apstats2 = ApertureStats(data, aperture, error=error, wcs=wcs,
                             sigma_clip=sigclip)

    unit = u.Jy
    apstats1_units = ApertureStats(data * u.Jy, aperture,
                                   error=error * u.Jy, wcs=wcs,
                                   sigma_clip=None)
    apstats2_units = ApertureStats(data * u.Jy, aperture,
                                   error=error * u.Jy, wcs=wcs,
                                   sigma_clip=sigclip)

    @pytest.mark.parametrize('with_units', [True, False])
    @pytest.mark.parametrize('with_sigmaclip', [True, False])
    def test_properties(self, with_units, with_sigmaclip):
        apstats = [self.apstats1.copy(), self.apstats2.copy(),
                   self.apstats1_units.copy(), self.apstats2_units.copy()]
        index = [1, 3] if with_sigmaclip else [0, 2]
        index = index[1] if with_units else index[0]
        apstats1 = apstats[index]
        apstats2 = apstats1.copy()

        idx = 1

        scalar_props = ('isscalar', 'n_apertures')

        # Evaluate (cache) properties before slice
        for prop in apstats1.properties:
            _ = getattr(apstats1, prop)
        apstats3 = apstats1[idx]
        for prop in apstats1.properties:
            if prop in scalar_props:
                continue
            assert_equal(getattr(apstats1, prop)[idx], getattr(apstats3, prop))

        # Slice catalog before evaluating catalog properties
        apstats4 = apstats2[idx]
        for prop in apstats1.properties:
            if prop in scalar_props:
                continue
            assert_equal(getattr(apstats4, prop), getattr(apstats1, prop)[idx])

    def test_skyaperture(self):
        pix_apstats = ApertureStats(self.data, self.aperture, wcs=self.wcs)
        skyaper = self.aperture.to_sky(self.wcs)
        sky_apstats = ApertureStats(self.data, skyaper, wcs=self.wcs)

        exclude_props = ('bbox', 'error_sum_cutout', 'sum_error',
                         'sky_centroid', 'sky_centroid_icrs')
        for prop in pix_apstats.properties:
            if prop in exclude_props:
                continue
            assert_allclose(getattr(pix_apstats, prop),
                            getattr(sky_apstats, prop), atol=1e-7)

        match = 'A wcs is required when using a SkyAperture'
        with pytest.raises(ValueError, match=match):
            _ = ApertureStats(self.data, skyaper)

    def test_lazyproperties_class_cache(self):
        """
        Test that _lazyproperties is cached on the class and shared
        across instances.
        """
        apstats2 = ApertureStats(self.data, self.aperture)
        result1 = self.apstats1._lazyproperties
        result2 = apstats2._lazyproperties
        assert result1 is result2

    def test_minimal_inputs(self):
        apstats = ApertureStats(self.data, self.aperture)
        props = ('sky_centroid', 'sky_centroid_icrs', 'error_sum_cutout')
        for prop in props:
            assert set(getattr(apstats, prop)) == {None}
        assert np.all(np.isnan(apstats.sum_err))
        assert set(apstats._variance_cutout) == {None}

        apstats = ApertureStats(self.data, self.aperture, sum_method='center')
        assert set(apstats._variance_cutout_center) == {None}

    @pytest.mark.parametrize('sum_method', ['exact', 'subpixel'])
    def test_sum_method(self, sum_method):
        apstats1 = ApertureStats(self.data, self.aperture, error=self.error,
                                 sum_method='center')
        apstats2 = ApertureStats(self.data, self.aperture, error=self.error,
                                 sum_method=sum_method, subpixels=4)

        scalar_props = ('isscalar', 'n_apertures')

        # Evaluate (cache) properties before slice
        for prop in apstats1.properties:
            if prop in scalar_props:
                continue
            if 'sum' in prop:
                # Test that these properties are not equal
                with pytest.raises(AssertionError):
                    assert_equal(getattr(apstats1, prop),
                                 getattr(apstats2, prop))
            else:
                assert_equal(getattr(apstats1, prop), getattr(apstats2, prop))

    def test_sum_method_photometry(self):
        for method in ('center', 'exact', 'subpixel'):
            subpixels = 4
            apstats = ApertureStats(self.data, self.aperture,
                                    error=self.error, sum_method=method,
                                    subpixels=subpixels)
            apsum, apsum_err = self.aperture.do_photometry(self.data,
                                                           error=self.error,
                                                           method=method,
                                                           subpixels=subpixels)
            assert_allclose(apstats.sum, apsum)
            assert_allclose(apstats.sum_err, apsum_err)

    def test_mask(self):
        mask = np.zeros(self.data.shape, dtype=bool)
        mask[225:240, 80:90] = True  # partially mask id=2
        mask[190:210, 40:60] = True  # completely mask id=3

        apstats = ApertureStats(self.data, self.aperture, mask=mask,
                                error=self.error)

        # id=2 is partially masked
        assert apstats[1].sum < self.apstats1[1].sum
        assert apstats[1].sum_err < self.apstats1[1].sum_err

        exclude = ('isscalar', 'n_apertures', 'sky_centroid',
                   'sky_centroid_icrs')
        apstats1 = apstats[2]
        for prop in apstats1.properties:
            if (prop in exclude or 'bbox' in prop or 'cutout' in prop
                    or 'moments' in prop):
                continue
            assert np.all(np.isnan(getattr(apstats1, prop)))

        # Test that mask=None is the same as mask=np.ma.nomask
        apstats1 = ApertureStats(self.data, self.aperture, mask=None)
        apstats2 = ApertureStats(self.data, self.aperture, mask=np.ma.nomask)
        assert_equal(apstats1.centroid, apstats2.centroid)

    def test_local_bkg(self):
        data = np.ones(self.data.shape) * 100.0
        local_bkg = (10, 20, 30)
        apstats = ApertureStats(data, self.aperture, local_bkg=local_bkg)

        for i, locbkg in enumerate(local_bkg):
            apstats0 = ApertureStats(data - locbkg, self.aperture[i],
                                     local_bkg=None)
            for prop in apstats.properties:
                assert_equal(getattr(apstats[i], prop),
                             getattr(apstats0, prop))

        # Test broadcasting
        local_bkg = (12, 12, 12)
        apstats1 = ApertureStats(data, self.aperture, local_bkg=local_bkg)
        apstats2 = ApertureStats(data, self.aperture, local_bkg=local_bkg[0])
        assert_equal(apstats1.sum, apstats2.sum)

        match = 'local_bkg must be scalar or have the same length as the'
        with pytest.raises(ValueError, match=match):
            _ = ApertureStats(data, self.aperture, local_bkg=(10, 20))

        match = 'local_bkg must not contain any non-finite'
        with pytest.raises(ValueError, match=match):
            _ = ApertureStats(data, self.aperture[0:2], local_bkg=(10, np.nan))
        with pytest.raises(ValueError, match=match):
            _ = ApertureStats(data, self.aperture[0:2],
                              local_bkg=(-np.inf, 10))

        match = 'local_bkg must be a 1D array'
        with pytest.raises(ValueError, match=match):
            _ = ApertureStats(data, self.aperture[0:2],
                              local_bkg=np.ones((3, 3)))

    def test_no_aperture_overlap(self):
        aperture = CircularAperture(((0, 0), (100, 100), (-100, -100)), r=5)
        apstats = ApertureStats(self.data, aperture)
        assert_equal(apstats._overlap, [True, True, False])

        exclude = ('isscalar', 'n_apertures', 'sky_centroid',
                   'sky_centroid_icrs')
        apstats1 = apstats[2]
        for prop in apstats1.properties:
            if (prop in exclude or 'bbox' in prop or 'cutout' in prop
                    or 'moments' in prop):
                continue
            assert np.all(np.isnan(getattr(apstats1, prop)))

    def test_to_table(self):
        tbl = self.apstats1.to_table()
        assert tbl.colnames == self.apstats1.default_columns
        assert len(tbl) == len(self.apstats1) == 3

        columns = ['id', 'min', 'max', 'mean', 'median', 'std', 'sum']
        tbl = self.apstats1.to_table(columns=columns)
        assert tbl.colnames == columns
        assert len(tbl) == len(self.apstats1) == 3

        tbl = self.apstats1.to_table(columns='sum')
        assert tbl.colnames == ['sum']
        assert len(tbl) == len(self.apstats1) == 3

    def test_slicing(self):
        apstats = self.apstats1
        _ = apstats.to_table()
        apstat0 = apstats[1]
        assert apstat0.n_apertures == 1
        assert apstat0.ids == np.array([2])
        apstat1 = apstats.select_id(2)
        assert apstat1.n_apertures == 1
        assert apstat0.sum_aper_area == apstat1.sum_aper_area

        apstat0 = apstats[0:1]
        assert len(apstat0) == 1

        apstat0 = apstats[0:2]
        assert len(apstat0) == 2

        apstat0 = apstats[0:3]
        assert len(apstat0) == 3

        apstat0 = apstats[1:]
        apstat1 = apstats.select_ids([1, 2])
        assert len(apstat0) == len(apstat1) == 2

        apstat0 = apstats[1:]
        apstat1 = apstats.select_ids([1, 2])
        assert len(apstat0) == len(apstat1) == 2

        apstat0 = apstats[[2, 1, 0]]
        apstat1 = apstats.select_ids([3, 2, 1])
        assert len(apstat0) == len(apstat1) == 3
        assert_equal(apstat0.ids, [3, 2, 1])
        assert_equal(apstat1.ids, [3, 2, 1])

        # Test select_ids when ids are not sorted
        apstat0 = apstats[[2, 1, 0]]
        apstat1 = apstat0.select_ids(2)
        assert apstat1.ids == 2

        mask = apstats.id >= 2
        apstat0 = apstats[mask]
        assert len(apstat0) == 2
        assert_equal(apstat0.ids, [2, 3])

        # Test iter
        for (i, apstat) in enumerate(apstats):
            assert apstat.isscalar
            assert apstat.id == (i + 1)

        match = "Scalar 'ApertureStats' object has no len"
        with pytest.raises(TypeError, match=match):
            _ = len(apstats[0])

        apstat0 = apstats[0]
        match = "A scalar 'ApertureStats' object cannot be indexed"
        with pytest.raises(TypeError, match=match):
            apstat1 = apstat0[0]
        apstat0 = apstats[0]
        with pytest.raises(TypeError, match=match):
            apstat1 = apstat0[0]  # can't slice scalar object

        match = '-1 is not a valid source ID number'
        with pytest.raises(ValueError, match=match):
            apstat0 = apstats.select_ids([-1, 0])

    def test_scalar_aperture_stats(self):
        apstats = self.apstats1[0]
        assert apstats.n_apertures == 1
        assert apstats.ids == np.array([1])
        tbl = apstats.to_table()
        assert len(tbl) == 1

    def test_deprecated_attributes(self):
        """
        Test that deprecated attributes are still available and
        give the same value as the new attributes, but raise an
        AstropyDeprecationWarning.
        """
        apstats = ApertureStats(self.data, self.aperture, error=self.error)
        match = 'attribute was deprecated'
        deprecated_map = {
            'covar_sigx2': 'covariance_xx',
            'covar_sigxy': 'covariance_xy',
            'covar_sigy2': 'covariance_yy',
            'cxx': 'ellipse_cxx',
            'cxy': 'ellipse_cxy',
            'cyy': 'ellipse_cyy',
            'data_sumcutout': 'data_sum_cutout',
            'error_sumcutout': 'error_sum_cutout',
            'get_id': 'select_id',
            'get_ids': 'select_ids',
            'semimajor_sigma': 'semimajor_axis',
            'semiminor_sigma': 'semiminor_axis',
            'xcentroid': 'x_centroid',
            'ycentroid': 'y_centroid',
        }
        for old_name, new_name in deprecated_map.items():
            with pytest.warns(AstropyDeprecationWarning, match=match):
                old_val = getattr(apstats, old_name)
            new_val = getattr(apstats, new_name)
            assert_equal(old_val, new_val)

    def test_invalid_inputs(self):
        match = 'aperture must be an Aperture or Region object'
        with pytest.raises(TypeError, match=match):
            ApertureStats(self.data, 10.0)

        with (patch.dict(sys.modules, {'regions': None}),
              pytest.raises(TypeError, match=match)):
            ApertureStats(self.data, 10.0)

        match = 'sigma_clip must be a SigmaClip instance'
        with pytest.raises(TypeError, match=match):
            ApertureStats(self.data, self.aperture, sigma_clip=10)

        match = 'error must be a 2D array'
        with pytest.raises(ValueError, match=match):
            ApertureStats(self.data, self.aperture, error=10.0)

        match = 'error must be a 2D array'
        with pytest.raises(ValueError, match=match):
            ApertureStats(self.data, self.aperture, error=np.ones(3))

        match = 'data and error must have the same shape'
        with pytest.raises(ValueError, match=match):
            ApertureStats(self.data, self.aperture, error=np.ones((3, 3)))

    def test_repr_str(self):
        assert repr(self.apstats1) == str(self.apstats1)
        assert 'Length: 3' in repr(self.apstats1)

    def test_data_dtype(self):
        """
        Regression test that input ``data`` with int dtype does not
        raise UFuncTypeError due to subtraction of float array from int
        array.
        """
        data = np.ones((25, 25), dtype=np.uint16)
        aper = CircularAperture((12, 12), 5)
        apstats = ApertureStats(data, aper)
        assert apstats.min == 1.0
        assert apstats.max == 1.0
        assert apstats.mean == 1.0
        assert apstats.x_centroid == 12.0
        assert apstats.y_centroid == 12.0

    @pytest.mark.parametrize('with_units', [True, False])
    def test_nddata_input(self, with_units):
        mask = np.zeros(self.data.shape, dtype=bool)
        mask[225:240, 80:90] = True  # partially mask id=2

        data = self.data
        error = self.error
        if with_units:
            unit = u.Jy
            data <<= unit
            error <<= unit
        else:
            unit = None

        apstats1 = ApertureStats(data, self.aperture, error=error,
                                 mask=mask, wcs=self.wcs, sigma_clip=None)

        uncertainty = StdDevUncertainty(self.error)
        nddata = NDData(self.data, uncertainty=uncertainty, mask=mask,
                        wcs=self.wcs, unit=unit)
        apstats2 = ApertureStats(nddata, self.aperture, sigma_clip=None)

        assert_allclose(apstats1.x_centroid, apstats2.x_centroid)
        assert_allclose(apstats1.y_centroid, apstats2.y_centroid)
        assert_allclose(apstats1.sum, apstats2.sum)

        if with_units:
            assert apstats1.sum.unit == unit

        match = 'keyword will be ignored'
        nddata = NDData(self.data, uncertainty=uncertainty, mask=mask,
                        wcs=self.wcs, unit=unit)
        with pytest.warns(AstropyUserWarning, match=match):
            ApertureStats(nddata, self.aperture, mask=mask)

    def test_tiny_source(self):
        data = np.zeros((21, 21))
        data[5, 5] = 1.0
        aperture = CircularAperture(((5, 5), (15, 15)), r=1)
        apstats = ApertureStats(data, aperture)
        assert_allclose(apstats.sum, (1.0, 0.0))
        assert_allclose(apstats[0].covariance,
                        [(1 / 12, 0), (0, 1 / 12)] * u.pix**2)
        assert_allclose(apstats[1].covariance,
                        [(np.nan, np.nan), (np.nan, np.nan)] * u.pix**2)
        assert_allclose(apstats.fwhm, [0.67977799, np.nan] * u.pix)


@pytest.mark.skipif(not HAS_REGIONS, reason='regions is required')
def test_aperture_stats_region():
    from regions import CirclePixelRegion, PixCoord
    region = CirclePixelRegion(center=PixCoord(5, 5), radius=3)
    aperture = CircularAperture((5, 5), r=3)
    data = np.ones((10, 10))
    apstats1 = ApertureStats(data, region)
    apstats2 = ApertureStats(data, aperture)
    tbl = apstats1.to_table()
    for colname in tbl.colnames:
        val1 = getattr(apstats1, colname)
        if val1 is not None:
            assert_allclose(val1, getattr(apstats2, colname))


def test_aperture_metadata():
    x = [10, 20, 3]
    y = [3, 5, 10]
    xypos = list(zip(x, y, strict=False))
    a1 = CircularAperture(xypos, r=3)
    a2 = CircularAperture(xypos, r=4)
    a3 = CircularAnnulus(xypos, 5, 10)
    a4 = EllipticalAperture(xypos, 10, 5, theta=10 * u.deg)
    a5 = EllipticalAnnulus(xypos, a_in=5, a_out=10, b_in=3, b_out=5,
                           theta=20 * u.deg)
    a6 = RectangularAperture(xypos, 10, 5, theta=30 * u.deg)
    a7 = RectangularAnnulus(xypos, w_in=5, w_out=10, h_in=3, h_out=5,
                            theta=40 * u.deg)
    apers = (a1, a2, a3, a4, a5, a6, a7)
    data = np.ones((50, 50))

    for aper in apers:
        apstats = ApertureStats(data, aper)
        tbl = apstats.to_table()
        assert tbl.meta['aperture'] == aper.__class__.__name__
        params = aper._params
        for param in params:
            if param != 'positions':
                assert tbl.meta[f'aperture_{param}'] == getattr(aper, param)

    wcs = make_wcs(data.shape)
    skycoord = wcs.pixel_to_world(10, 10)
    unit = u.arcsec
    saper = SkyEllipticalAnnulus(skycoord, a_in=0.1 * unit, a_out=0.2 * unit,
                                 b_in=0.05 * unit, b_out=0.1 * unit,
                                 theta=10 * u.deg)
    apstats = ApertureStats(data, saper, wcs=wcs)
    tbl = apstats.to_table()
    assert tbl.meta['aperture'] == saper.__class__.__name__
    assert tbl.meta['aperture_a_in'] == saper.a_in
    assert tbl.meta['aperture_a_out'] == saper.a_out
    assert tbl.meta['aperture_b_out'] == saper.b_out
    assert tbl.meta['aperture_theta'] == saper.theta



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the finder module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.convolution import convolve
from astropy.modeling.models import Gaussian2D
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.datasets import make_100gaussians_image
from photutils.segmentation.finder import SourceFinder
from photutils.segmentation.utils import make_2dgaussian_kernel
from photutils.utils._optional_deps import HAS_SKIMAGE
from photutils.utils.exceptions import NoDetectionsWarning


class TestSourceFinder:
    data = make_100gaussians_image() - 5.0  # subtract background
    kernel = make_2dgaussian_kernel(3.0, size=5)
    convolved_data = convolve(data, kernel, normalize_kernel=True)
    threshold = 1.5 * 2.0
    n_pixels = 10

    @pytest.mark.skipif(not HAS_SKIMAGE, reason='skimage is required')
    def test_deblend(self):
        """
        Test deblend.
        """
        finder = SourceFinder(n_pixels=self.n_pixels, progress_bar=False)
        segm1 = finder(self.convolved_data, self.threshold)
        assert segm1.n_labels == 94

        segm2 = finder(self.convolved_data << u.uJy, self.threshold * u.uJy)
        assert segm2.n_labels == 94
        assert np.all(segm1.data == segm2.data)

    def test_invalid_units(self):
        """
        Test invalid units.
        """
        finder = SourceFinder(n_pixels=self.n_pixels, progress_bar=False)
        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            finder(self.convolved_data << u.uJy, self.threshold)
        with pytest.raises(ValueError, match=match):
            finder(self.convolved_data, self.threshold * u.uJy)
        with pytest.raises(ValueError, match=match):
            finder(self.convolved_data << u.uJy, self.threshold * u.m)

    def test_no_deblend(self):
        """
        Test no deblend.
        """
        finder = SourceFinder(n_pixels=self.n_pixels, deblend=False,
                              progress_bar=False)
        segm = finder(self.convolved_data, self.threshold)
        assert segm.n_labels == 87

    def test_no_sources(self):
        """
        Test no sources.
        """
        finder = SourceFinder(n_pixels=self.n_pixels, deblend=True,
                              progress_bar=False)

        match = 'No sources were found'
        with pytest.warns(NoDetectionsWarning, match=match):
            segm = finder(self.convolved_data, 1000)
        assert segm is None

    @pytest.mark.skipif(not HAS_SKIMAGE, reason='skimage is required')
    def test_n_pixels_tuple(self):
        """
        Test n_pixels tuple.
        """
        g1 = Gaussian2D(10, 35, 45, 5, 5)
        g2 = Gaussian2D(10, 50, 50, 5, 5)
        g3 = Gaussian2D(10, 66, 55, 5, 5)
        yy, xx = np.mgrid[0:101, 0:101]
        data = g1(xx, yy) + g2(xx, yy) + g3(xx, yy)

        sf1 = SourceFinder(n_pixels=200)
        segm1 = sf1(data, threshold=0.1)
        assert segm1.n_labels == 1

        sf2 = SourceFinder(n_pixels=(200, 5))
        segm2 = sf2(data, threshold=0.1)
        assert segm2.n_labels == 3

    def test_repr(self):
        """
        Test repr.
        """
        finder = SourceFinder(n_pixels=self.n_pixels, deblend=False,
                              progress_bar=False)
        cls_repr = repr(finder)
        assert cls_repr.startswith(finder.__class__.__name__)


def test_finder_deprecations():
    finder = SourceFinder(n_pixels=10, progress_bar=False)
    match = 'attribute was deprecated'
    with pytest.warns(AstropyDeprecationWarning, match=match):
        _ = finder.npixels
    with pytest.warns(AstropyDeprecationWarning, match=match):
        _ = finder.nlevels



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Photutils is an Astropy affiliated package to provide tools for
detecting and performing photometry of astronomical sources.

It also has tools for background estimation, ePSF building, PSF
matching, radial profiles, centroiding, and morphological measurements.
"""

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''

from .utils._deprecation import use_future_column_names  # noqa: F401


future_column_names = False
"""
If `True`, all photutils functions return standard
`~astropy.table.QTable` (or `~astropy.table.Table`) instances with the
new column names instead of deprecated-column subclasses. Set this to
`True` after updating your code to use the new column names to verify
compatibility with the 4.0 behavior.

For a scoped override that does not affect the global flag, use
`photutils.use_future_column_names` as a context manager::

    with photutils.use_future_column_names():
        table = cat.to_table()  # returns a plain QTable
"""



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _components module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.fitting import LevMarLSQFitter, TRFLSQFitter
from astropy.table import QTable, Table
from numpy.testing import assert_equal

from photutils.detection import DAOStarFinder
from photutils.psf import CircularGaussianPRF
from photutils.psf._components import (PSFDataProcessor, PSFFitter,
                                       PSFResultsAssembler)
from photutils.psf.photometry import _PSFParameterMapper


@pytest.fixture(name='basic_data')
def fixture_basic_data():
    """
    Create basic test data for component testing.
    """
    # Create simple test image
    shape = (25, 25)
    yy, xx = np.mgrid[:shape[0], :shape[1]]

    # Add a simple Gaussian source
    model = CircularGaussianPRF(flux=100, x_0=12, y_0=12, fwhm=2.0)
    data = model(xx, yy)
    rng = np.random.default_rng(0)
    data += rng.normal(0, 1, shape)

    error = np.ones_like(data)
    mask = np.zeros_like(data, dtype=bool)

    return data, error, mask


@pytest.fixture(name='psf_model')
def fixture_psf_model():
    """
    Create a basic PSF model for testing.
    """
    return CircularGaussianPRF(flux=1, fwhm=2.7)


@pytest.fixture(name='param_mapper')
def fixture_param_mapper(psf_model):
    """
    Create a parameter mapper for testing.
    """
    return _PSFParameterMapper(psf_model)


@pytest.fixture(name='init_params')
def fixture_init_params():
    """
    Create initial parameters table for testing.
    """
    return Table({
        'x_init': [12.0, 8.0, 16.0],
        'y_init': [12.0, 8.0, 16.0],
        'flux_init': [100.0, 50.0, 75.0],
    })


class TestPSFDataProcessor:
    """
    Test the PSFDataProcessor class.
    """

    def test_init(self, param_mapper):
        """
        Test PSFDataProcessor initialization.
        """
        fit_shape = (7, 7)
        processor = PSFDataProcessor(
            param_mapper=param_mapper,
            fit_shape=fit_shape,
            finder=None,
            aperture_radius=3.0,
            local_bkg_estimator=None,
        )

        assert processor.param_mapper is param_mapper
        assert processor.fit_shape == fit_shape
        assert processor.finder is None
        assert processor.aperture_radius == 3.0
        assert processor.local_bkg_estimator is None
        assert processor.data_unit is None
        assert processor.finder_results is None
        assert processor._cached_offsets is None
        assert processor._cache_key is None

    def test_validate_array_valid_input(self, param_mapper):
        """
        Test validate_array with valid inputs.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))

        # Test valid 2D array
        data = np.ones((10, 10))
        result = processor.validate_array(data, 'data')
        assert_equal(result, data)

        # Test None input
        result = processor.validate_array(None, 'mask')
        assert result is None

        # Test np.ma.nomask
        result = processor.validate_array(np.ma.nomask, 'mask')
        assert result is None

    def test_validate_array_invalid_input(self, param_mapper):
        """
        Test validate_array with invalid inputs.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))

        # Test 1D array
        match = 'data must be a 2D array'
        with pytest.raises(ValueError, match=match):
            processor.validate_array(np.ones(10), 'data')

        # Test 3D array
        match = 'error must be a 2D array'
        with pytest.raises(ValueError, match=match):
            processor.validate_array(np.ones((5, 5, 5)), 'error')

        # Test shape mismatch
        data = np.ones((10, 10))
        error = np.ones((5, 5))
        match_str = 'data and error must have the same shape'
        with pytest.raises(ValueError, match=match_str):
            processor.validate_array(error, 'error', data_shape=data.shape)

    def test_normalize_init_units_no_units(self, param_mapper):
        """
        Test normalize_init_units when neither have units.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))
        processor.data_unit = None

        init_params = Table({'flux_init': [100.0, 200.0]})
        result = processor.normalize_init_units(init_params, 'flux_init')

        assert result is init_params
        assert result['flux_init'].unit is None

    def test_normalize_init_units_both_have_units(self, param_mapper):
        """
        Test normalize_init_units when both have compatible units.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))
        processor.data_unit = u.count

        init_params = QTable({'flux_init': [100.0, 200.0] * u.count})
        result = processor.normalize_init_units(init_params, 'flux_init')

        assert result is init_params
        assert result['flux_init'].unit == u.count

    def test_normalize_init_units_incompatible_units(self, param_mapper):
        """
        Test normalize_init_units with incompatible units.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))
        processor.data_unit = u.count

        init_params = QTable({'flux_init': [100.0, 200.0] * u.meter})
        match_str = 'incompatible with the input data units'
        with pytest.raises(ValueError, match=match_str):
            processor.normalize_init_units(init_params, 'flux_init')

    def test_normalize_init_units_data_has_units_init_does_not(
            self, param_mapper):
        """
        Test normalize_init_units when data has units but init doesn't.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))
        processor.data_unit = u.count

        init_params = Table({'flux_init': [100.0, 200.0]})
        match_str = 'input data has units.*does not have units'
        with pytest.raises(ValueError, match=match_str):
            processor.normalize_init_units(init_params, 'flux_init')

    def test_normalize_init_units_init_has_units_data_does_not(
            self, param_mapper):
        """
        Test normalize_init_units when init has units, but data doesn't.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))
        processor.data_unit = None

        init_params = QTable({'flux_init': [100.0, 200.0] * u.count})
        match_str = 'has units.*input data does not have units'
        with pytest.raises(ValueError, match=match_str):
            processor.normalize_init_units(init_params, 'flux_init')

    def test_validate_init_params_valid(self, param_mapper, init_params):
        """
        Test validate_init_params with valid input.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))
        processor.data_unit = None

        result = processor.validate_init_params(init_params)

        assert isinstance(result, Table)
        assert 'x_init' in result.colnames
        assert 'y_init' in result.colnames
        assert 'flux_init' in result.colnames

    def test_validate_init_params_none(self, param_mapper):
        """
        Test validate_init_params with None input.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))

        result = processor.validate_init_params(None)
        assert result is None

    def test_validate_init_params_invalid_type(self, param_mapper):
        """
        Test validate_init_params with invalid type.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))

        match_str = 'init_params must be an astropy Table'
        with pytest.raises(TypeError, match=match_str):
            processor.validate_init_params([1, 2, 3])

    def test_validate_init_params_missing_positions(self, param_mapper):
        """
        Test validate_init_params with missing position columns.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))

        # Missing x column
        init_params = Table({'y_init': [12.0], 'flux_init': [100.0]})
        match_str = 'must contain valid column names.*x and y'
        with pytest.raises(ValueError, match=match_str):
            processor.validate_init_params(init_params)

        # Missing y column
        init_params = Table({'x_init': [12.0], 'flux_init': [100.0]})
        with pytest.raises(ValueError, match=match_str):
            processor.validate_init_params(init_params)

    def test_validate_init_params_nonfinite_local_bkg(self, param_mapper):
        """
        Test validate_init_params allows non-finite local_bkg values.

        Non-finite local_bkg values should be allowed and will be
        flagged later during processing.
        """
        processor = PSFDataProcessor(param_mapper, (7, 7))
        processor.data_unit = None

        init_params = Table({
            'x_init': [12.0],
            'y_init': [12.0],
            'local_bkg': [np.nan],
        })

        # Should not raise an error - non-finite local_bkg is allowed
        result = processor.validate_init_params(init_params)
        assert result is not None
        assert 'local_bkg' in result.colnames
        assert np.isnan(result['local_bkg'][0])

    def test_get_aper_fluxes(self, param_mapper, basic_data, init_params):
        """
        Test get_aper_fluxes method.
        """
        data, _, mask = basic_data
        processor = PSFDataProcessor(param_mapper, (7, 7), aperture_radius=3.0)

        fluxes = processor.get_aper_fluxes(data, mask, init_params)

        assert isinstance(fluxes, np.ndarray)
        assert len(fluxes) == len(init_params)
        assert np.all(np.isfinite(fluxes))

    def test_find_sources_if_needed_with_init_params(self, param_mapper,
                                                     basic_data, init_params):
        """
        Test find_sources_if_needed when init_params is provided.
        """
        data, _, mask = basic_data
        processor = PSFDataProcessor(param_mapper, (7, 7))

        result = processor.find_sources_if_needed(data, mask, init_params)

        assert result is not None
        assert 'id' in result.colnames
        assert len(result) == len(init_params)

    def test_find_sources_if_needed_no_init_no_finder(self, param_mapper,
                                                      basic_data):
        """
        Test find_sources_if_needed when neither is provided.
        """
        data, _, mask = basic_data
        processor = PSFDataProcessor(param_mapper, (7, 7), finder=None)

        match_str = 'finder must be defined if init_params is not input'
        with pytest.raises(ValueError, match=match_str):
            processor.find_sources_if_needed(data, mask, None)

    def test_find_sources_if_needed_with_finder(self, param_mapper,
                                                basic_data):
        """
        Test find_sources_if_needed with a finder.
        """
        data, _, mask = basic_data
        finder = DAOStarFinder(threshold=5.0, fwhm=2.0)
        processor = PSFDataProcessor(param_mapper, (7, 7), finder=finder)

        result = processor.find_sources_if_needed(data, mask, None)

        # Should find at least one source in our test data
        assert result is not None
        assert len(result) >= 1


class TestPSFFitter:
    """
    Test the PSFFitter class.
    """

    def test_init(self, psf_model, param_mapper):
        """
        Test PSFFitter initialization.
        """
        fitter = PSFFitter(
            psf_model=psf_model,
            param_mapper=param_mapper,
            fitter=None,
            fitter_maxiters=100,
            xy_bounds=None,
            group_warning_threshold=25,
        )

        assert fitter.psf_model is psf_model
        assert fitter.param_mapper is param_mapper
        assert isinstance(fitter.fitter, TRFLSQFitter)  # Default fitter
        assert fitter.fitter_maxiters == 100
        assert fitter.xy_bounds is None
        assert fitter.group_warning_threshold == 25

    def test_init_custom_fitter(self, psf_model, param_mapper):
        """
        Test PSFFitter initialization with custom fitter.
        """
        custom_fitter = LevMarLSQFitter()
        fitter = PSFFitter(
            psf_model=psf_model,
            param_mapper=param_mapper,
            fitter=custom_fitter,
            xy_bounds=(1.0, 1.0),
        )

        assert fitter.fitter is custom_fitter
        assert fitter.xy_bounds == (1.0, 1.0)

    def test_make_psf_model_single_source(self, psf_model, param_mapper):
        """
        Test make_psf_model with a single source.
        """
        fitter = PSFFitter(psf_model, param_mapper)

        sources = Table({
            'id': [1],
            'x_init': [10.0],
            'y_init': [15.0],
            'flux_init': [100.0],
        })

        model = fitter.make_psf_model(sources)

        assert model.name == 1
        assert model.x_0.value == 10.0
        assert model.y_0.value == 15.0
        assert model.flux.value == 100.0

    def test_make_psf_model_multiple_sources(self, psf_model, param_mapper):
        """
        Test make_psf_model with multiple sources.
        """
        fitter = PSFFitter(psf_model, param_mapper)

        sources = Table({
            'id': [1, 2],
            'x_init': [10.0, 20.0],
            'y_init': [15.0, 25.0],
            'flux_init': [100.0, 200.0],
        })

        model = fitter.make_psf_model(sources)

        # Should be a flat model with parameters for each source
        assert hasattr(model, 'flux_0')
        assert hasattr(model, 'x_0_0')
        assert hasattr(model, 'y_0_0')
        assert hasattr(model, 'flux_1')
        assert hasattr(model, 'x_0_1')
        assert hasattr(model, 'y_0_1')

        # Check parameter values
        assert model.flux_0.value == 100.0
        assert model.x_0_0.value == 10.0
        assert model.y_0_0.value == 15.0
        assert model.flux_1.value == 200.0
        assert model.x_0_1.value == 20.0
        assert model.y_0_1.value == 25.0

    def test_make_psf_model_with_xy_bounds(self, psf_model, param_mapper):
        """
        Test make_psf_model with xy bounds.
        """
        fitter = PSFFitter(psf_model, param_mapper, xy_bounds=(2.0, 3.0))

        sources = Table({
            'id': [1],
            'x_init': [10.0],
            'y_init': [15.0],
            'flux_init': [100.0],
        })

        model = fitter.make_psf_model(sources)

        # Check bounds were set
        assert model.x_0.bounds == (8.0, 12.0)  # 10.0 ± 2.0
        assert model.y_0.bounds == (12.0, 18.0)  # 15.0 ± 3.0

    def test_make_psf_model_with_units(self, psf_model, param_mapper):
        """
        Test make_psf_model with quantity columns.
        """
        fitter = PSFFitter(psf_model, param_mapper)

        sources = QTable({
            'id': [1],
            'x_init': [10.0] * u.pixel,
            'y_init': [15.0] * u.pixel,
            'flux_init': [100.0] * u.count,
        })

        model = fitter.make_psf_model(sources)

        # Units should be stripped for fitting
        assert model.x_0.value == 10.0
        assert model.y_0.value == 15.0
        assert model.flux.value == 100.0


class TestPSFResultsAssembler:
    """
    Test the PSFResultsAssembler class.
    """

    def test_init(self, param_mapper):
        """
        Test PSFResultsAssembler initialization.
        """
        fit_shape = (7, 7)
        assembler = PSFResultsAssembler(
            param_mapper=param_mapper,
            fit_shape=fit_shape,
            xy_bounds=(1.0, 1.0),
        )

        assert assembler.param_mapper is param_mapper
        assert assembler.fit_shape == fit_shape
        assert assembler.xy_bounds == (1.0, 1.0)

    def test_get_fit_error_indices_all_converged(self, param_mapper):
        """
        Test get_fit_error_indices when all fits converged.
        """
        assembler = PSFResultsAssembler(param_mapper, (7, 7))

        # Good convergence status codes
        fit_info = [
            {'ierr': 1},  # Converged
            {'status': 2},  # Converged
            {'ierr': 3},  # Converged
        ]

        bad_indices = assembler.get_fit_error_indices(fit_info)

        assert len(bad_indices) == 0

    def test_get_fit_error_indices_some_failed(self, param_mapper):
        """
        Test get_fit_error_indices when some fits failed.
        """
        assembler = PSFResultsAssembler(param_mapper, (7, 7))

        fit_info = [
            {'ierr': 1},  # Converged
            {'ierr': 0},  # Failed
            {'status': 2},  # Converged
            {'status': 0},  # Failed
        ]

        bad_indices = assembler.get_fit_error_indices(fit_info)

        assert_equal(bad_indices, [1, 3])

    def test_param_errors_to_table_no_units(self, param_mapper):
        """
        Test param_errors_to_table without units.
        """
        assembler = PSFResultsAssembler(param_mapper, (7, 7))

        # Assuming 3 fitted parameters (x, y, flux)
        fit_param_errs = np.array([
            [0.1, 0.1, 1.0],  # errors for source 1
            [0.2, 0.2, 2.0],  # errors for source 2
        ])

        table = assembler.param_errors_to_table(fit_param_errs, data_unit=None)

        assert isinstance(table, QTable)
        assert 'x_err' in table.colnames
        assert 'y_err' in table.colnames
        assert 'flux_err' in table.colnames
        assert len(table) == 2

    def test_param_errors_to_table_with_units(self, param_mapper):
        """
        Test param_errors_to_table with units.
        """
        assembler = PSFResultsAssembler(param_mapper, (7, 7))

        fit_param_errs = np.array([
            [0.1, 0.1, 1.0],
            [0.2, 0.2, 2.0],
        ])

        table = assembler.param_errors_to_table(fit_param_errs,
                                                data_unit=u.count)

        assert isinstance(table, QTable)
        assert table['flux_err'].unit == u.count
        # Position errors should not have units
        assert (not hasattr(table['x_err'], 'unit')
                or table['x_err'].unit is None)


class TestComponentIntegration:
    """
    Test integration between components.
    """

    def test_components_work_together(self, psf_model, basic_data,
                                      init_params):
        """
        Test that all components can work together in sequence.
        """
        data, _, mask = basic_data

        # Create components
        param_mapper = _PSFParameterMapper(psf_model)
        processor = PSFDataProcessor(param_mapper, (7, 7), aperture_radius=3.0)
        fitter = PSFFitter(psf_model, param_mapper)
        assembler = PSFResultsAssembler(param_mapper, (7, 7))

        # Process data
        processor.data_unit = None
        validated_params = processor.validate_init_params(init_params)
        sources = processor.find_sources_if_needed(data, mask,
                                                   validated_params)

        assert sources is not None
        assert len(sources) > 0

        # Create PSF model for fitting (test with first source)
        psf_model_fit = fitter.make_psf_model(sources[:1])

        assert psf_model_fit is not None
        assert hasattr(psf_model_fit, 'x_0')
        assert hasattr(psf_model_fit, 'y_0')

        # Test error handling for fit results
        fit_info = [{'ierr': 1}]  # Successful fit
        bad_indices = assembler.get_fit_error_indices(fit_info)
        assert len(bad_indices) == 0

    def test_components_preserve_units(self, psf_model):
        """
        Test that components properly handle astropy units.
        """
        # Create data with units
        data = np.ones((25, 25)) * u.count
        mask = np.zeros((25, 25), dtype=bool)

        init_params = QTable({
            'x_init': [12.0] * u.pixel,
            'y_init': [12.0] * u.pixel,
            'flux_init': [100.0] * u.count,
        })

        # Create components
        param_mapper = _PSFParameterMapper(psf_model)
        processor = PSFDataProcessor(param_mapper, (7, 7))

        # Set data unit and validate
        processor.data_unit = u.count
        validated_params = processor.validate_init_params(init_params)
        sources_with_id = processor.find_sources_if_needed(
            data, mask, validated_params)

        assert sources_with_id['flux_init'].unit == u.count

        # Test that PSF fitting strips units
        fitter = PSFFitter(psf_model, param_mapper)
        psf_model_fit = fitter.make_psf_model(sources_with_id)

        # Model parameters should be unitless for fitting
        assert not hasattr(psf_model_fit.x_0.value, 'unit')
        assert not hasattr(psf_model_fit.y_0.value, 'unit')
        assert not hasattr(psf_model_fit.flux.value, 'unit')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for centroiding sources using Gaussians.
"""

import warnings

import numpy as np
from astropy.modeling.fitting import TRFLSQFitter
from astropy.modeling.models import Gaussian1D, Gaussian2D
from astropy.utils.exceptions import AstropyUserWarning

from photutils.centroids._utils import (_gaussian1d_moments,
                                        _gaussian2d_moments,
                                        _validate_gaussian_inputs)
from photutils.utils._deprecation import deprecated_positional_kwargs
from photutils.utils._quantity_helpers import process_quantities

__all__ = ['centroid_1dg', 'centroid_2dg']


@deprecated_positional_kwargs(since='3.0', until='4.0')
def centroid_1dg(data, error=None, mask=None):
    """
    Calculate the centroid of a 2D array by fitting 1D Gaussians to the
    marginal ``x`` and ``y`` distributions of the array.

    Non-finite values (e.g., NaN or inf) in the ``data`` or ``error``
    arrays are automatically masked. The final mask is a logical OR
    combination of the input ``mask``, the automatically generated mask
    for non-finite values, and the mask of the input ``data`` if it is a
    `~numpy.ma.MaskedArray`. The centroid is calculated using only the
    unmasked data values.

    Parameters
    ----------
    data : 2D array_like
        The 2D image data. ``data`` can be a `~numpy.ma.MaskedArray`.
        The image should be a background-subtracted cutout image
        containing a single source.

    error : 2D `~numpy.ndarray`, optional
        The 2D array of the 1-sigma errors of the input ``data``.

    mask : 2D bool `~numpy.ndarray`, optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        If ``data`` is a `~numpy.ma.MaskedArray`, its mask will be
        combined (using bitwise OR) with the input ``mask``.

    Returns
    -------
    centroid : `~numpy.ndarray`
        The ``x, y`` coordinates of the centroid.

    Examples
    --------
    >>> import numpy as np
    >>> from photutils.datasets import make_4gaussians_image
    >>> from photutils.centroids import centroid_1dg
    >>> data = make_4gaussians_image()
    >>> data -= np.median(data[0:30, 0:125])
    >>> data = data[40:80, 70:110]
    >>> x1, y1 = centroid_1dg(data)
    >>> print(np.array((x1, y1)))
    [19.96553246 20.04952841]

    .. plot::

        import matplotlib.pyplot as plt
        import numpy as np
        from photutils.centroids import centroid_1dg
        from photutils.datasets import make_4gaussians_image

        data = make_4gaussians_image()
        data -= np.median(data[0:30, 0:125])
        data = data[40:80, 70:110]
        xycen = centroid_1dg(data)
        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
        ax.imshow(data, origin='lower')
        ax.scatter(*xycen, color='red', marker='+', s=100, label='Centroid')
        ax.legend()
    """
    (data, error), _ = process_quantities((data, error), ('data', 'error'))
    data, mask, error = _validate_gaussian_inputs(data, mask, error)

    if error is not None:
        error_squared = error**2
        xy_error = [np.sqrt(np.sum(error_squared, axis=i)) for i in (0, 1)]
        xy_weights = [1.0 / xy_err.clip(min=1.0e-30) for xy_err in xy_error]
    else:
        xy_weights = [np.ones(data.shape[i]) for i in (1, 0)]

    # Assign zero weight where an entire row or column is masked
    if np.any(mask):
        bad_idx = [np.all(mask, axis=i) for i in (0, 1)]
        for i in (0, 1):
            xy_weights[i][bad_idx[i]] = 0.0

    xy_data = [np.sum(data, axis=i) for i in (0, 1)]

    # Gaussian1D stddev is bounded to be strictly positive
    fitter = TRFLSQFitter()

    centroid = []
    for (data_i, weights_i) in zip(xy_data, xy_weights, strict=True):
        params_init = _gaussian1d_moments(data_i)
        g_init = Gaussian1D(*params_init)
        x = np.arange(data_i.size)
        g_fit = fitter(g_init, x, data_i, weights=weights_i)
        centroid.append(g_fit.mean.value)

    return np.array(centroid)


@deprecated_positional_kwargs(since='3.0', until='4.0')
def centroid_2dg(data, error=None, mask=None):
    """
    Calculate the centroid of a 2D array by fitting a 2D Gaussian to the
    array.

    Non-finite values (e.g., NaN or inf) in the ``data`` or ``error``
    arrays are automatically masked. The final mask is a logical OR
    combination of the input ``mask``, the automatically generated mask
    for non-finite values, and the mask of the input ``data`` if it is a
    `~numpy.ma.MaskedArray`. The centroid is calculated using only the
    unmasked data values.

    Parameters
    ----------
    data : 2D array_like
        The 2D image data. ``data`` can be a `~numpy.ma.MaskedArray`.
        The image should be a background-subtracted cutout image
        containing a single source.

    error : 2D `~numpy.ndarray`, optional
        The 2D array of the 1-sigma errors of the input ``data``.

    mask : 2D bool `~numpy.ndarray`, optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        If ``data`` is a `~numpy.ma.MaskedArray`, its mask will be
        combined (using bitwise OR) with the input ``mask``.

    Returns
    -------
    centroid : `~numpy.ndarray`
        The ``x, y`` coordinates of the centroid.

    Examples
    --------
    >>> import numpy as np
    >>> from photutils.datasets import make_4gaussians_image
    >>> from photutils.centroids import centroid_2dg
    >>> data = make_4gaussians_image()
    >>> data -= np.median(data[0:30, 0:125])
    >>> data = data[40:80, 70:110]
    >>> x1, y1 = centroid_2dg(data)
    >>> print(np.array((x1, y1)))
    [19.9851944  20.01490157]

    .. plot::

        import matplotlib.pyplot as plt
        import numpy as np
        from photutils.centroids import centroid_2dg
        from photutils.datasets import make_4gaussians_image

        data = make_4gaussians_image()
        data -= np.median(data[0:30, 0:125])
        data = data[40:80, 70:110]
        xycen = centroid_2dg(data)
        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
        ax.imshow(data, origin='lower')
        ax.scatter(*xycen, color='red', marker='+', s=100, label='Centroid')
        ax.legend()
    """
    (data, error), _ = process_quantities((data, error), ('data', 'error'))
    data, mask, error = _validate_gaussian_inputs(data, mask, error)

    if np.count_nonzero(~mask) < 6:
        msg = ('Input data must have a least 6 unmasked values to fit a '
               '2D Gaussian.')
        raise ValueError(msg)

    # Subtract the minimum of the data to make the data values positive.
    # Moments from negative data values can yield undefined Gaussian
    # parameters, e.g., x_stddev and y_stddev.
    shifted = data - np.min(data)
    if np.sum(shifted) == 0:
        msg = ('Input data must have non-constant values to fit a '
               '2D Gaussian.')
        raise ValueError(msg)

    if error is not None:
        weights = 1.0 / error.clip(min=1.0e-30)
    else:
        weights = np.ones(data.shape)

    # Assign zero weight to masked pixels
    if np.any(mask):
        weights[mask] = 0.0

    amplitude, x_mean, y_mean, x_stddev, y_stddev, theta = _gaussian2d_moments(
        shifted)

    g_init = Gaussian2D(amplitude=amplitude,
                        x_mean=x_mean,
                        y_mean=y_mean,
                        x_stddev=x_stddev,
                        y_stddev=y_stddev,
                        theta=theta)
    fitter = TRFLSQFitter()

    y, x = np.indices(data.shape)

    with warnings.catch_warnings(record=True) as fit_warnings:
        warnings.simplefilter('always', AstropyUserWarning)
        gfit = fitter(g_init, x, y, data, weights=weights)

    if any(issubclass(w.category, AstropyUserWarning) for w in fit_warnings):
        msg = 'The fit may not have converged. Please check your results.'
        warnings.warn(msg, AstropyUserWarning)

    return np.array([gfit.x_mean.value, gfit.y_mean.value])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the local_background module.
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose

from photutils.aperture import CircularAnnulus
from photutils.background import LocalBackground, MedianBackground


def test_local_background_invalid_radii():
    """
    Test that LocalBackground raises errors for invalid radius values.
    """
    # Test negative inner radius
    match = 'inner_radius must be positive'
    with pytest.raises(ValueError, match=match):
        LocalBackground(-5, 10)

    # Test zero inner radius
    with pytest.raises(ValueError, match=match):
        LocalBackground(0, 10)

    # Test negative outer radius
    match = 'outer_radius must be positive'
    with pytest.raises(ValueError, match=match):
        LocalBackground(5, -10)

    # Test zero outer radius
    with pytest.raises(ValueError, match=match):
        LocalBackground(5, 0)

    # Test outer_radius <= inner_radius
    match = 'outer_radius must be greater than inner_radius'
    with pytest.raises(ValueError, match=match):
        LocalBackground(10, 5)

    # Test equal radii
    with pytest.raises(ValueError, match=match):
        LocalBackground(10, 10)


def test_local_background():
    """
    Test the basic functionality of LocalBackground with a simple
    constant data array.
    """
    data = np.ones((101, 101))
    local_bkg = LocalBackground(5, 10, bkg_estimator=MedianBackground())

    x = np.arange(1, 7) * 10
    y = np.arange(1, 7) * 10
    bkg = local_bkg(data, x, y)
    assert_allclose(bkg, np.ones(len(x)))

    # Test scalar x and y
    bkg2 = local_bkg(data, x[2], y[2])
    assert not isinstance(bkg2, np.ndarray)
    assert_allclose(bkg[2], bkg2)

    bkg3 = local_bkg(data, -100, -100)
    assert np.isnan(bkg3)

    match = "'positions' must not contain any non-finite"
    with pytest.raises(ValueError, match=match):
        _ = local_bkg(data, x[2], np.inf)

    cls_repr = repr(local_bkg)
    assert cls_repr.startswith(local_bkg.__class__.__name__)

    # Test default bkg_estimator
    local_bkg2 = LocalBackground(5, 10, bkg_estimator=None)
    bkg4 = local_bkg2(data, x, y)
    assert_allclose(bkg4, bkg)


def test_local_background_estimator_1d():
    """
    Test that the bkg_estimator can be a 1D function that takes an array
    and returns a scalar.
    """

    def estimator(data):
        assert data.ndim == 1
        return np.nanmedian(data)

    data = np.ones((51, 51))
    local_bkg = LocalBackground(3, 6, bkg_estimator=estimator)
    bkg = local_bkg(data, [10, 20], [10, 20])
    assert_allclose(bkg, np.ones(2))


def test_to_aperture_scalar():
    """
    Test to_aperture method with scalar x and y positions.
    """
    r_in = 5
    r_out = 10
    local_bkg = LocalBackground(r_in, r_out)

    # Test scalar positions
    x = 50.0
    y = 50.0
    aperture = local_bkg.to_aperture(x, y)

    # Check aperture type and properties
    assert isinstance(aperture, CircularAnnulus)
    assert_allclose(aperture.positions, [[x, y]])
    assert_allclose(aperture.r_in, r_in)
    assert_allclose(aperture.r_out, r_out)


def test_to_aperture_array():
    """
    Test to_aperture method with array x and y positions.
    """
    r_in = 7.5
    r_out = 15.2
    local_bkg = LocalBackground(r_in, r_out)

    # Test array positions
    x = np.array([10.0, 20.1, 35.3])
    y = np.array([14.4, 27.2, 33.4])
    xypos = list(zip(x, y, strict=False))
    aperture = local_bkg.to_aperture(x, y)

    # Check aperture type and properties
    assert isinstance(aperture, CircularAnnulus)
    assert_allclose(aperture.positions, xypos)
    assert_allclose(aperture.r_in, r_in)
    assert_allclose(aperture.r_out, r_out)

    # Test list positions
    x = list(x)
    y = list(y)
    aperture2 = local_bkg.to_aperture(x, y)
    assert aperture == aperture2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for deprecation warnings when optional arguments are passed
positionally.
"""

import numpy as np
import pytest
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.segmentation.core import SegmentationImage
from photutils.segmentation.utils import make_2dgaussian_kernel


class TestSegmentationImagePositionalKwargs:
    """
    Test SegmentationImage methods warn for positional optional args.
    """

    def setup_method(self):
        self.data = np.array([[1, 1, 0, 0, 2, 2],
                              [1, 1, 0, 0, 2, 2],
                              [0, 0, 3, 3, 0, 0],
                              [0, 0, 3, 3, 0, 0]])

    def test_reassign_label_positional_warns(self):
        segm = SegmentationImage(self.data.copy())
        match = 'reassign_label'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            segm.reassign_label(1, 4, True)  # noqa: FBT003

    def test_reassign_label_keyword_no_warning(self):
        segm = SegmentationImage(self.data.copy())
        segm.reassign_label(1, 4, relabel=True)

    def test_keep_label_positional_warns(self):
        segm = SegmentationImage(self.data.copy())
        match = 'keep_label'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            segm.keep_label(1, True)  # noqa: FBT003

    def test_keep_label_keyword_no_warning(self):
        segm = SegmentationImage(self.data.copy())
        segm.keep_label(1, relabel=True)

    def test_remove_label_positional_warns(self):
        segm = SegmentationImage(self.data.copy())
        match = 'remove_label'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            segm.remove_label(1, True)  # noqa: FBT003

    def test_remove_label_keyword_no_warning(self):
        segm = SegmentationImage(self.data.copy())
        segm.remove_label(1, relabel=True)

    def test_remove_border_labels_positional_warns(self):
        segm = SegmentationImage(self.data.copy())
        match = 'remove_border_labels'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            segm.remove_border_labels(1, True)  # noqa: FBT003

    def test_remove_border_labels_keyword_no_warning(self):
        segm = SegmentationImage(self.data.copy())
        segm.remove_border_labels(1, partial_overlap=True)

    def test_remove_masked_labels_positional_warns(self):
        segm = SegmentationImage(self.data.copy())
        mask = np.zeros(self.data.shape, dtype=bool)
        mask[0, 0] = True
        match = 'remove_masked_labels'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            segm.remove_masked_labels(mask, True)  # noqa: FBT003

    def test_remove_masked_labels_keyword_no_warning(self):
        segm = SegmentationImage(self.data.copy())
        mask = np.zeros(self.data.shape, dtype=bool)
        mask[0, 0] = True
        segm.remove_masked_labels(mask, partial_overlap=True)

    def test_make_cutout_positional_warns(self):
        segm = SegmentationImage(self.data.copy())
        segment = segm.segments[0]
        data = np.random.default_rng(0).random(self.data.shape)
        match = 'make_cutout'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            segment.make_cutout(data, True)  # noqa: FBT003

    def test_make_cutout_keyword_no_warning(self):
        segm = SegmentationImage(self.data.copy())
        segment = segm.segments[0]
        data = np.random.default_rng(0).random(self.data.shape)
        segment.make_cutout(data, masked_array=True)


class TestMake2DGaussianKernelPositionalKwargs:
    """
    Test make_2dgaussian_kernel warns for positional optional args.
    """

    def test_positional_warns(self):
        match = 'make_2dgaussian_kernel'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            make_2dgaussian_kernel(3.0, 5, 'oversample')

    def test_keyword_no_warning(self):
        make_2dgaussian_kernel(3.0, 5, mode='oversample')
